import asyncio
import csv
import psutil
import uuid
import os
import functools
import signal
import time
from typing import List, TextIO, Dict, Union
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric.metrics_set import MetricsSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from .csv_config import CSVConfig
has_connector = True


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop, 
    events_file: TextIO
): 
    try:
        events_file.close()
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass
    

class CSV:

    def __init__(self, config: CSVConfig) -> None:
        self.events_filepath = config.events_filepath
        self.metrics_filepath = Path(config.metrics_filepath).absolute()
        self.experiments_filepath = config.experiments_filepath
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        
        experiments_path = Path(self.experiments_filepath)
        experiments_directory = experiments_path.parent
        experiments_filename = experiments_path.stem

        self.variants_filepath = os.path.join(
            experiments_directory,
            f'{experiments_filename}_variants.csv'
        )

        self.mutations_filepath = os.path.join(
            experiments_directory,
            f'{experiments_filename}_mutations.csv'
        )

        self.streams_filepath = config.streams_filepath

        system_metrics_path = Path(config.system_metrics_filepath)
        system_metrics_directory = system_metrics_path.parent
        system_metrics_filename = system_metrics_path.stem

        self.stage_system_metrics_filepath = os.path.join(
            system_metrics_directory,
            f'{system_metrics_filename}_stages.csv'
        )

        self.session_system_metrics_filepath = os.path.join(
            system_metrics_directory,
            f'{system_metrics_filename}_session.csv'
        )

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._events_csv_writer: csv.DictWriter = None
        self._metrics_csv_writer: csv.DictWriter = None
        self._stage_metrics_csv_writer: csv.DictWriter = None
        self._errors_csv_writer: csv.DictWriter = None
        self._custom_metrics_csv_writer: csv.DictWriter = None
        self._experiments_writer: csv.DictWriter = None
        self._variants_writer: csv.DictWriter = None
        self._mutations_writer: csv.DictWriter = None
        self._streams_writer: csv.DictWriter = None
        self._stage_system_metrics_writer: csv.DictWriter = None
        self._session_system_metrics_writer: csv.DictWriter = None

        self._loop: asyncio.AbstractEventLoop = None

        self.events_file: TextIO = None 
        self.metrics_file: TextIO = None
        self.experiments_file: TextIO = None
        self.variants_file: TextIO = None
        self.mutations_file: TextIO = None
        self.streams_file: TextIO = None
        self.stage_system_metrics_file: TextIO = None
        self.session_system_metrics_file: TextIO = None

        self.write_mode = 'w' if config.overwrite else 'a'


        filepath = Path(config.metrics_filepath)
        base_filepath = filepath.parent
        base_filename = filepath.stem

        self.shared_metrics_filepath = os.path.join(
            base_filepath,
            f'{base_filename}_shared.csv'
        )

        self.custom_metrics_filepath = os.path.join(
            base_filepath,
            f'{base_filename}_custom.csv'
        )

        self.errors_metrics_filepath = os.path.join(
            base_filepath,
            f'{base_filename}_errors.csv'
        )

    async def connect(self):
        self._loop = asyncio._get_running_loop()
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Setting filepaths')
        
        original_filepath = Path(self.events_filepath)
        
        directory = original_filepath.parent
        filename = original_filepath.stem

        events_file_timestamp = time.time()

        self.events_filepath = os.path.join(
            directory,
            f'{filename}_{events_file_timestamp}.csv'
        )

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):
        if self.session_system_metrics_file is None:
            self.session_system_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.session_system_metrics_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.session_system_metrics_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Session System Metrics to file - {self.session_system_metrics_filepath}')

        if self._session_system_metrics_writer is None or self.write_mode == 'w':
            self._session_system_metrics_writer = csv.DictWriter(
                self.session_system_metrics_file, 
                fieldnames=[
                    header for header in SystemMetricsSet.metrics_header_keys if header != 'stage'
                ]
            )

            await self._loop.run_in_executor(
                self._executor,
                self._session_system_metrics_writer.writeheader
            )

        metrics_sets: List[Dict[str, Union[int, float, str]]] = []

        for metrics_set in system_metrics_sets:
                
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics.record)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics.record)

        await self._loop.run_in_executor(
            self._executor,
            self._session_system_metrics_writer.writerows,
            metrics_sets
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Session System Metrics to file - {self.session_system_metrics_filepath}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        if self.stage_system_metrics_file is None:
            self.stage_system_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.stage_system_metrics_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.stage_system_metrics_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Stage System Metrics to file - {self.stage_system_metrics_filepath}')

        if self._stage_system_metrics_writer is None or self.write_mode == 'w':
            self._stage_metrics_csv_writer = csv.DictWriter(
                self.stage_system_metrics_file, 
                fieldnames=SystemMetricsSet.metrics_header_keys
            )

            await self._loop.run_in_executor(
                self._executor,
                self._stage_metrics_csv_writer.writeheader
            )

        metrics_sets: List[Dict[str, Union[int, float, str]]] = []

        for metrics_set in system_metrics_sets:

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics.record)

        await self._loop.run_in_executor(
            self._executor,
            self._stage_metrics_csv_writer.writerows,
            metrics_sets
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Stage System Metrics to file - {self.stage_system_metrics_filepath}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):
        if self.streams_file is None:
            self.streams_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.streams_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.streams_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Streams to file - {self.streams_filepath}')

        streams_data: List[Dict[str, Union[float, int]]] = []

        for stream_name, stream in stream_metrics.items():

            streams_data.extend([{
                    'stage': stream_name,
                    'group': group_name,
                    **group_metrics
            } for group_name, group_metrics in stream.grouped.items()])
            

        headers = list(streams_data[0].keys())

        if self._streams_writer is None or self.write_mode == 'w':
            self._streams_writer = csv.DictWriter(
                self.streams_file, 
                fieldnames=headers
            )

            await self._loop.run_in_executor(
                self._executor,
                self._streams_writer.writeheader
            )

        await self._loop.run_in_executor(
            self._executor,
            self._streams_writer.writerows,
            streams_data
        )

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Experiments to file - {self.experiments_filepath}')

        if self.experiments_file is None:
            self.experiments_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.experiments_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.experiments_file
                    )
                )

        if self._experiments_writer is None or self.write_mode == 'w':
            self._experiments_writer = csv.DictWriter(
                self.experiments_file, 
                fieldnames=experiment_metrics.experiments_metrics_fields
            )

            await self._loop.run_in_executor(
                self._executor,
                self._experiments_writer.writeheader
            )

        await self._loop.run_in_executor(
            self._executor,
            self._experiments_writer.writerows,
            experiment_metrics.experiments
        )

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        if self.variants_file is None:
            self.variants_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.variants_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.variants_file
                    )
                )
        
        if self._variants_writer is None or self.write_mode == 'w':
            self._variants_writer = csv.DictWriter(
                self.variants_file, 
                fieldnames=experiment_metrics.variants_metrics_fields
            )

            await self._loop.run_in_executor(
                self._executor,
                self._variants_writer.writeheader
            )

        await self._loop.run_in_executor(
            self._executor,
            self._variants_writer.writerows,
            experiment_metrics.variants
        )

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        if self.mutations_file is None:
            self.mutations_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.mutations_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.mutations_file
                    )
                )  

        if self._mutations_writer is None or self.write_mode == 'w':
            self._mutations_writer = csv.DictWriter(
                self.mutations_file, 
                fieldnames=experiment_metrics.mutations_metrics_fields
            )

            await self._loop.run_in_executor(
                self._executor,
                self._mutations_writer.writeheader
            )

        await self._loop.run_in_executor(
            self._executor,
            self._mutations_writer.writerows,
            experiment_metrics.mutations
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Experiments to file - {self.experiments_filepath}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Events to file - {self.events_filepath}')

        if self.events_file is None:
            self.events_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.events_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.events_file
                    )
                )
          
        for event in events:
            if self._events_csv_writer is None or self.write_mode == 'w':
                self._events_csv_writer = csv.DictWriter(self.events_file, fieldnames=event.fields)
                
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(event.fields)}')

                await self._loop.run_in_executor(
                    self._executor,
                    self._events_csv_writer.writeheader
                )

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(event.fields)}')

            await self._loop.run_in_executor(
                self._executor,
                self._events_csv_writer.writerow,
                event.record
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Events to file - {self.events_filepath}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Shared Metrics to file - {self.metrics_filepath}')

        headers = [
            'name',
            'stage',
            'group',
            'total',
            'succeeded',
            'failed',
            'actions_per_second'
        ]

        shared_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.shared_metrics_filepath,
                    'w'
                )
        )

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop,
                    shared_metrics_file
                )
            )

        if self._stage_metrics_csv_writer is None:
            self._stage_metrics_csv_writer = csv.DictWriter(shared_metrics_file, fieldnames=headers)


            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(headers)}')

            await self._loop.run_in_executor(
                self._executor,
                self._stage_metrics_csv_writer.writeheader
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(headers)}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self._loop.run_in_executor(
                self._executor,
                self._stage_metrics_csv_writer.writerow,
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }
            )

        await self._loop.run_in_executor(
            self._executor,
            shared_metrics_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Shared Metrics to file - {self.metrics_filepath}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Metrics to file - {self.metrics_filepath}')

        metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                self.metrics_filepath, 
                'w'
            )
        )

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop,
                    metrics_file
                )
            )

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._metrics_csv_writer is None:

                headers = [
                    *metrics_set.fields,
                    'group'
                ]

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(headers)}')

                self._metrics_csv_writer = csv.DictWriter(metrics_file, fieldnames=headers)

                await self._loop.run_in_executor(
                    self._executor,
                    self._metrics_csv_writer.writeheader
                )

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(headers)}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')

                await self._loop.run_in_executor(
                    self._executor,
                    self._metrics_csv_writer.writerow,
                    {
                        **group.record,
                        'group': group_name
                    }
                )

        await self._loop.run_in_executor(
            self._executor,
            metrics_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Metrics to file - {self.metrics_filepath}')
        
    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Custom Metrics to file - {self.metrics_filepath}')

        custom_metrics_file = None
        
        headers = [
            'name',
            'stage',
            'group',
        ]

        if self._custom_metrics_csv_writer is None:

            custom_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.custom_metrics_filepath,
                    'w'
                )       
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        custom_metrics_file
                    )
                )

            for metrics_set in metrics_sets:
                for custom_metric in metrics_set.custom_metrics.values():
                    headers.append(
                        custom_metric.metric_name
                    )


            self._custom_metrics_csv_writer = csv.DictWriter(custom_metrics_file, fieldnames=headers)
            
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(headers)}')

            await self._loop.run_in_executor(
                self._executor,
                self._custom_metrics_csv_writer.writeheader
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(headers)}')


        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - Custom')

            await self._loop.run_in_executor(
                self._executor,
                self._custom_metrics_csv_writer.writerow,
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'custom',
                    **{
                        custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                    }
                }
            )

        await self._loop.run_in_executor(
            self._executor,
            custom_metrics_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Custom Metrics to file - {self.metrics_filepath}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Error Metrics to file - {self.metrics_filepath}')

        error_csv_headers = [
            'name',
            'stage',
            'error_message',
            'error_count'
        ]

        errors_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                self.errors_metrics_filepath,
                'w'
            )
        )

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop,
                    errors_file
                )
            )
            
        if self._errors_csv_writer is None:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(error_csv_headers)}')
            error_csv_writer = csv.DictWriter(errors_file, fieldnames=error_csv_headers)

            await self._loop.run_in_executor(
                self._executor,
                error_csv_writer.writeheader
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(error_csv_headers)}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                await self._loop.run_in_executor(
                    self._executor,
                    error_csv_writer.writerow,
                    {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    }
                )

        await self._loop.run_in_executor(
            self._executor,
            errors_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Error Metrics to file - {self.metrics_filepath}')
                
    async def close(self):
        
        if self.events_file:
            await self._loop.run_in_executor(
                self._executor,
                self.events_file.close
            )

        if self.experiments_file:
            await self._loop.run_in_executor(
                self._executor,
                self.experiments_file.close
            )

        if self.variants_file:
            await self._loop.run_in_executor(
                self._executor,
                self.variants_file.close
            )

        if self.mutations_file:
            await self._loop.run_in_executor(
                self._executor,
                self.mutations_file.close
            )

        if self.streams_file:
            await self._loop.run_in_executor(
                self._executor,
                self.streams_file.close
            )

        if self.stage_system_metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                self.stage_system_metrics_file.close
            )

        if self.session_system_metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                self.session_system_metrics_file.close
            )

        self._executor.shutdown(wait=False, cancel_futures=True)
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')




import math
import numpy
from typing import Union, List, Tuple, Optional
from numpy.random import normal

class SciPyDistribution:

    def rvs(self, size: int):
        return NotImplemented('Err. - This class is a generalized definition')


class BaseDistribution:

    def __init__(
        self, 
        size: Union[int,float]=1000,
        center: Union[int,float]=0.5,
        randomness: Union[int, float]=0.25,
        frozen_distribution: SciPyDistribution=None
    ) -> None:
        self.size = size
        self.center = center
        self.randomness = randomness
        self._frozen_distribution = frozen_distribution
        self._noise_scale = 0.01
        self._lower_bound = 0.1
        self._upper_bound = 0.9

    def generate_distribution(self, batch_size: int) -> List[float]:
        return [
            math.ceil(
                step_batch_size
            ) for step_batch_size in self._generate_distribution(
                scale_factor=batch_size
            )
        ]
    
    def generate_non_scaled_distribution(self):
        return self._generate_distribution()
    
    def _generate_distribution(self, scale_factor: Optional[int]=None):
        distribution_size, step_size = self._get_distribution_and_step_size()

        distribution_sample = self._frozen_distribution.rvs(
            distribution_size
        ) * normal(
            scale=self._noise_scale,
            size=distribution_size
        )

        generated_walk = self._generate_random_walk(distribution_sample)

        smoothed_walk = self._smooth_generated_walk(
            generated_walk,
            step_size
        )

        if scale_factor:
            smoothed_walk *= scale_factor
        
        return self._apply_averaging(
            smoothed_walk,
            step_size
        )
    
    
    def _get_distribution_and_step_size(self) -> Tuple[int, int]:
        if self.size <= 10:
            distribution_size = self.size**3
            step_size = self.size**2

        elif self.size < 100:
            distribution_size = self.size**2
            step_size = self.size

        elif self.size < 1000:
            distribution_size = self.size * int(math.sqrt(self.size))
            step_size = int(math.sqrt(self.size))
            
        else:
            distribution_size = self.size * (1 + int(
                math.sqrt(self.size)/self.size
            ))

            step_size = (1 + int(
                math.sqrt(self.size)/self.size
            ))

        return distribution_size, step_size
    
    def _generate_random_walk(self, distribution_sample: List[float]) -> List[float]:
        current_step_value = 0.5
        result = []
        for sample_value in distribution_sample:

            if current_step_value < self._lower_bound:
                current_step_value = self._lower_bound

            elif current_step_value > self._upper_bound:
                current_step_value = self._upper_bound

            result.append(current_step_value)
            current_step_value += sample_value

        return numpy.array(result)
    
    def _smooth_generated_walk(
        self, 
        generated_walk: List[float], 
        smoothing_window_size: int
    ) -> List[float]:
        return numpy.convolve(
            generated_walk, 
            numpy.ones(
                (smoothing_window_size,)
            )/smoothing_window_size
        )[(smoothing_window_size-1):]
    
    def _apply_averaging(
        self, 
        scaled_walk: List[float],
        step_size: int
    ) -> List[float]:
        averaged_data = []

        for idx in range(0, len(scaled_walk), step_size):
            averaged_data.append(
                numpy.mean(scaled_walk[idx:idx + self.size])
            )

        return averaged_data


from .act import Act
from .analyze import Analyze
from .execute import Execute
from .optimize import Optimize
from .setup import Setup
from .complete import Complete
from .submit import Submit
from .error import Error
from .idle import Idle


def load_domain_name(
    buffer: bytes, 
    offset: int
):
    parts = []
    cursor = None
    data_len = len(buffer)
    visited = set()

    while offset < data_len:

        if offset in visited:
            raise Exception(buffer, offset, 'Pointer loop detected')
        
        visited.add(offset)
        length = buffer[offset]
        offset += 1

        if length == 0:

            if cursor is None:
                cursor = offset

            break

        if length >= 0xc0:

            if cursor is None:
                cursor = offset + 1

            offset = (length - 0xc0) * 256 + buffer[offset]

            continue

        parts.append(buffer[offset:offset + length])
        offset += length

    if cursor is None:
        raise Exception(buffer, offset, 'Bad data')
    
    data = b'.'.join(parts).decode()

    return cursor, data


import inspect
from typing import Any, Callable, Awaitable, Optional, Tuple
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook


class ContextHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str,
        call: Callable[..., Awaitable[Any]], 
        *names:  Optional[Tuple[str, ...]],
        store: Optional[str]=None,
        load: Optional[str]=None,
        pre: bool=False,
        order: int=1,
        skip: bool=False
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.CONTEXT
        )
        
        self.names = list(set(names))
        self.store = store
        self.load = load
        self.pre = pre
        self.order = order

        self.args = inspect.signature(call)
        self.params = self.args.parameters

    async def call(self, **kwargs) -> None:

        if self.skip:
            return kwargs

        hook_args = {name: value for name, value in kwargs.items() if name in self.params}

        hook_args = hook_args
  
        for param_name in self.params.keys():

            context_value = self.context[param_name]

            if param_name != 'self' and context_value is not None:
                hook_args[param_name] = context_value
        
        if 'results' in list(self.params.keys()):
            results = []

            for stage_results in context_value.values():
                results.extend(
                    stage_results.results
                )

            context_value['results'] = results
        
        context_result = await self._call(**{name: value for name, value in hook_args.items() if name in self.params})

        if isinstance(context_result, dict):

            for context_key, value in context_result.items():
                self.context[context_key] = value

            return {
                **kwargs,
                **context_result
            }

        self.context[self.name] = context_result
        return {
            **kwargs,
            self.shortname: context_result
        }

    def copy(self):
        context_hook = ContextHook(
            self.name,
            self.shortname,
            self._call,
            *self.names,
            store=self.store,
            load=self.load,
            pre=self.pre,
            order=self.order,
            skip=self.skip,
        )

        context_hook.stage = self.stage

        return context_hook


from typing import Tuple, Optional
from pydantic import (
    BaseModel, 
    Field, 
    StrictStr, 
    StrictInt,
    StrictBool
)


class SaveHookValidator(BaseModel):
    names: Optional[Tuple[StrictStr, ...]]
    save_path: StrictStr=Field(..., min_length=1)
    order: StrictInt
    skip: StrictBool



import threading
import os
import time
import signal
import dill
import pickle
import warnings
import asyncio
import gc
from collections import defaultdict
from typing import (
    Dict, 
    Any, 
    List, 
    Union,
    Type
)
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.playwright import MercuryPlaywrightClient, ContextConfig
from hedra.core.engines.types.registry import RequestTypes
from hedra.core.engines.types.registry import registered_engines
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.personas import get_persona
from hedra.core.personas.persona_registry import registered_personas
from hedra.core.hooks.types.base.event_graph import EventGraph
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.load.hook import LoadHook
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.graphs.stages.base.parallel.partition_method import PartitionMethod
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.exceptions.process_killed_error import ProcessKilledError
from hedra.core.graphs.stages.base.import_tools import (
    import_stages, 
    import_plugins, 
    set_stage_hooks
)
from hedra.core.graphs.stages.setup.setup import Setup
from hedra.data.serializers import Serializer
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)
from hedra.plugins.extensions import get_enabled_extensions
from hedra.plugins.types.extension.types import ExtensionType
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.extension.extension_plugin import ExtensionPlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.reporting.reporter import ReporterConfig
from hedra.versioning.flags.types.base.active import active_flags
from hedra.versioning.flags.types.base.flag_type import FlagTypes
warnings.simplefilter("ignore")
warnings.filterwarnings("ignore")


async def start_execution(
    metadata_string: str=None,
    persona_config: Config=None,
    setup_stage: Setup=None,
    workers: int=None,
    worker_id: int=None,
    source_stage_name: str=None,
    source_stage_stream_configs: List[ReporterConfig]=[],
    logfiles_directory: str=None,
    log_level: str=None,
    extensions: Dict[str, ExtensionPlugin]={},
    loaded_actions: List[str]=[]
) -> Dict[str, Any]:

    current_task = asyncio.current_task()

    logging_manager.disable(
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.DISTRIBUTED_FILESYSTEM,
        LoggerTypes.SPINNER
    )

    logging_manager.update_log_level(log_level)
    logging_manager.logfiles_directory = logfiles_directory


    logger = HedraLogger()
    logger.logger_directory = logfiles_directory
    logger.initialize()
    await logger.filesystem.aio.create_logfile('hedra.core.log')
    logger.filesystem.create_filelogger('hedra.core.log')

    start = time.monotonic()
    
    await setup_stage.run_internal()

    stages: Dict[str, Stage] =  setup_stage.context['setup_stage_ready_stages']

    setup_execute_stage: Stage = stages.get(source_stage_name)
    setup_execute_stage.logger = logger

    persona = get_persona(persona_config)
    persona.stream_reporter_configs = source_stage_stream_configs
    persona.workers = workers

    actions_and_tasks: List[Union[ActionHook, TaskHook]] = []

    serializer = Serializer()
    if len(loaded_actions) > 0:
        for serialized_action in loaded_actions:
            actions_and_tasks.append(
                serializer.deserialize_action(serialized_action)
            )

    actions_and_tasks.extend(
        setup_stage.context['execute_stage_setup_hooks'].get(source_stage_name, [])
    )

    execution_hooks_count = len(actions_and_tasks)
    await logger.filesystem.aio['hedra.core'].info(
        f'{metadata_string} - Executing {execution_hooks_count} actions with a batch size of {persona_config.batch_size} for {persona_config.total_time} seconds using Persona - {persona.type.capitalize()}'
    )

    for hook in actions_and_tasks:

        if hook.action.type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

            await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Setting up Playwright Session')

            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')

            config = hook.session.config
            if config is None:
                config = ContextConfig(
                    browser_type=persona_config.browser_type,
                    device_type=persona_config.device_type,
                    locale=persona_config.locale,
                    geolocation=persona_config.geolocation,
                    permissions=persona_config.permissions,
                    color_scheme=persona_config.color_scheme,
                    options=persona_config.playwright_options
                )

            await hook.session.setup(
                config=config
            )

    for load_hook in setup_execute_stage.hooks[HookType.LOAD]:
        load_hook: LoadHook = load_hook
        load_hook.parser_config = setup_stage.config
    
    pipeline_stages = {
        setup_execute_stage.name: setup_execute_stage
    }

    hooks_by_type: Dict[
        HookType, 
        Union[
            List[ActionHook], 
            List[TaskHook]
        ]
    ] = defaultdict(list)

    for hook in actions_and_tasks:
        hooks_by_type[hook.hook_type].append(hook)
                
    persona.setup(hooks_by_type, metadata_string)

    persona.cpu_monitor.stage_type = StageTypes.EXECUTE
    persona.memory_monitor.stage_type = StageTypes.EXECUTE

    await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Starting execution')

    results = await persona.execute()

    elapsed = time.monotonic() - start

    await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Execution complete - Time (including addtional setup) took: {round(elapsed, 2)} seconds')

    context = {}

    for stage in pipeline_stages.values():

        for key, value in stage.context.as_serializable():    
            try:
                dill.dumps(value)
            
            except ValueError:
                stage.context.ignore_serialization_filters.append(key)
            
            except TypeError:
                stage.context.ignore_serialization_filters.append(key)

            except pickle.PicklingError:
                stage.context.ignore_serialization_filters.append(key)

        serializable_context = stage.context.as_serializable()

        context.update({
            context_key: context_value for context_key, context_value in serializable_context
        }) 

    for idx, result in enumerate(results):
        results[idx] = dill.dumps(result)  


    results_dict =  {
        'worker_idx': worker_id,
        'streamed_analytics': persona.streamed_analytics,
        'results': results,
        'total_results': len(results),
        'total_elapsed': persona.total_elapsed,
        'context': context,
        'monitoring': {
            'memory': persona.memory_monitor.collected,
            'cpu': persona.cpu_monitor.collected
        }
    }

    pending_tasks = [
        task for task in asyncio.all_tasks() if task != current_task
    ]
    
    for task in pending_tasks:
        if not task.cancelled():
            task.cancel()

    return results_dict


def execute_actions(parallel_config: str):

    import asyncio
    import uvloop
    uvloop.install()

    from hedra.logging import (
        logging_manager
    )

    try:
        loop = asyncio.get_event_loop()
    except Exception:

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    def handle_loop_stop(signame):
        try:
            pending_events = asyncio.all_tasks()
            
            for task in pending_events:
                if not task.cancelled():
                    try:
                        task.cancel()
                    except asyncio.CancelledError:
                        pass
                    except asyncio.InvalidStateError:
                        pass
                    
            loop.close()
            gc.collect()

        except BrokenPipeError:
            pass

        except RuntimeError:
            pass

    for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
        loop.add_signal_handler(
            getattr(signal, signame),
            lambda signame=signame: handle_loop_stop(signame)
        )

    try:
        parallel_config: Dict[str, Any] = dill.loads(parallel_config)
        
        graph_name = parallel_config.get('graph_name')
        graph_path: str= parallel_config.get('graph_path') 
        graph_id = parallel_config.get('graph_id')
        enable_unstable_features = parallel_config.get('enable_unstable_features', False)
    
        logfiles_directory = parallel_config.get('logfiles_directory')
        log_level = parallel_config.get('log_level')
        source_stage_name = parallel_config.get('source_stage_name')
        source_stage_context: Dict[str, Any] = parallel_config.get('source_stage_context')
        source_stage_plugins = parallel_config.get('source_stage_plugins')
        source_stage_config: Config = parallel_config.get('source_stage_config')
        source_stage_id = parallel_config.get('source_stage_id')
        source_setup_stage_name = parallel_config.get('source_setup_stage_name')
        source_stage_stream_configs = parallel_config.get('source_stage_stream_configs')
        source_stage_loaded_actions = parallel_config.get('source_stage_loaded_actions')
        partition_method = parallel_config.get('partition_method')
        worker_id = parallel_config.get('worker_id')
        workers = parallel_config.get('workers')

        thread_id = threading.current_thread().ident
        process_id = os.getpid()

        active_flags[FlagTypes.UNSTABLE_FEATURE] = enable_unstable_features

        logging_manager.disable(
            LoggerTypes.DISTRIBUTED,
            LoggerTypes.DISTRIBUTED_FILESYSTEM,
            LoggerTypes.SPINNER
        )

        logging_manager.update_log_level(log_level)
        logging_manager.logfiles_directory = logfiles_directory

        metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '


        discovered: Dict[str, Stage] = import_stages(graph_path)
        plugins_by_type = import_plugins(graph_path)

        initialized_stages = {}
        hooks_by_type = defaultdict(dict)
        hooks_by_name = {}
        hooks_by_shortname = defaultdict(dict)

        generated_hooks = {}
        for stage in discovered.values():
            stage: Stage = stage()

            if stage.context is None:
                stage.context = SimpleContext()
                
            stage.graph_name = graph_name
            stage.graph_path = graph_path
            stage.graph_id = graph_id

            for hook_shortname, hook in registrar.reserved[stage.name].items():
                hook._call = hook._call.__get__(stage, stage.__class__)
                setattr(stage, hook_shortname, hook._call)

            initialized_stage = set_stage_hooks(
                stage, 
                generated_hooks
            )

            for hook_type in initialized_stage.hooks:

                for hook in initialized_stage.hooks[hook_type]:
                    hook.stage_instance = initialized_stage
                    hooks_by_type[hook_type][hook.name] = hook
                    hooks_by_name[hook.name] = hook
                    hooks_by_shortname[hook_type][hook.shortname] = hook

            initialized_stages[initialized_stage.name] = initialized_stage

        execute_stage: Stage = initialized_stages.get(source_stage_name)
        execute_stage.context.update(source_stage_context)

        setup_stage: Setup = initialized_stages.get(source_setup_stage_name)
        setup_stage.context.update(source_stage_context)

        if partition_method == PartitionMethod.BATCHES and source_stage_config.optimized is False:
            if workers == worker_id:
                source_stage_config.batch_size = int(source_stage_config.batch_size/workers) + (source_stage_config.batch_size%workers)

                if source_stage_config.experiment:
                    distribution = source_stage_config.experiment.get('distribution')

                    for idx, distribution_value in enumerate(distribution):
                        distribution[idx] = int(distribution_value/workers) + (distribution_value%workers)

                    source_stage_config.experiment['distribution'] = distribution
            
            else:
                source_stage_config.batch_size = int(source_stage_config.batch_size/workers)

                if source_stage_config.experiment:
                    distribution = source_stage_config.experiment.get('distribution')

                    for idx, distribution_value in enumerate(distribution):
                        distribution[idx] = int(distribution_value/workers)

                    source_stage_config.experiment['distribution'] = distribution

        stage_persona_plugins: List[str] = source_stage_plugins[PluginType.PERSONA]
        persona_plugins: Dict[str, Type[PersonaPlugin]] = plugins_by_type[PluginType.PERSONA]

        stage_engine_plugins: List[str] = source_stage_plugins[PluginType.ENGINE]
        engine_plugins: Dict[str, Type[EnginePlugin]] = plugins_by_type[PluginType.ENGINE]
        
        stage_extension_plugins: List[str] = source_stage_plugins[PluginType.EXTENSION]
        extension_plugins: Dict[str, Type[ExtensionPlugin]] = get_enabled_extensions()

        enabled_extensions: Dict[str, ExtensionPlugin] = {}
        
        for plugin_name in stage_persona_plugins:
            plugin = persona_plugins.get(plugin_name)
            plugin.name = plugin_name
            registered_personas[plugin_name] = lambda config: plugin(config)
        
        for plugin_name in stage_engine_plugins:
            plugin = engine_plugins.get(plugin_name)
            plugin.name = plugin_name
            registered_engines[plugin_name] = lambda config: plugin(config)

        for plugin_name in stage_extension_plugins:
            plugin = extension_plugins.get(plugin_name)

            if plugin:
                enabled_plugin = plugin()
                enabled_extensions[enabled_plugin.name] = enabled_plugin

        events_graph = EventGraph(hooks_by_type)
        events_graph.hooks_to_events().assemble_graph().apply_graph_to_events()

        for stage in initialized_stages.values():
            stage.dispatcher.assemble_action_and_task_subgraphs()

        if setup_stage.context is None:
            setup_stage.context = SimpleContext()

        setup_stage.config = source_stage_config
        setup_stage.generation_setup_candidates = 1
        setup_stage.config = source_stage_config
        setup_stage.context['setup_stage_is_primary_thread'] = False
        setup_stage.context['setup_stage_target_config'] = source_stage_config
        setup_stage.context['setup_stage_target_stages'] = {
            execute_stage.name: execute_stage
        }

        results = loop.run_until_complete(
            start_execution(
                metadata_string=metadata_string,
                persona_config=source_stage_config,
                setup_stage=setup_stage,
                workers=workers,
                worker_id=worker_id,
                source_stage_name=source_stage_name,
                source_stage_stream_configs=source_stage_stream_configs,
                logfiles_directory=logfiles_directory,
                log_level=log_level,
                extensions=enabled_extensions,
                loaded_actions=source_stage_loaded_actions
            )
        )


        loop.close()
        gc.collect()

        return results

    except BrokenPipeError:
        raise ProcessKilledError()
    
    except RuntimeError:
        raise ProcessKilledError()

    except Exception as e:
        raise e


from typing import Dict, Union
from hedra.core.engines.client.config import Config
from .param_type import ParamType


class Batch:

    __slots__ = (
        'gradient',
        'size',
        'interval',
        'deferred'
    )

    def __init__(self, config: Config) -> None:
        self.gradient = config.batch_gradient

        self.size = config.batch_size

        self.interval = config.batch_interval

        self.deferred = []

    def to_params(self):
        return {
            'batch_gradient': {
                'value': self.gradient,
                'type': ParamType.FLOAT
            },
            'batch_size': {
                'value': self.size,
                'type': ParamType.INTEGER
            },
            'batch_interval': {
                'value': self.size,
                'type': ParamType.FLOAT
            }
        }

    def update(self, values: Dict[str, Union[int, float]]):
        gradient = values.get('batch_gradient')
        if gradient is None:
            gradient = self.gradient

        size = values.get('batch_size')
        if size is None:
            size = self.size

        interval = values.get('batch_interval')
        if interval is None:
            interval = self.interval

        self.gradient = gradient
        self.size = size
        self.interval = interval



from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import BaseModel, StrictStr


class BigTableConnectorConfig(BaseModel):
    service_account_json_path: StrictStr
    instance_id: StrictStr
    table_name: StrictStr
    connector_type: ConnectorType=ConnectorType.BigTable




from enum import Enum
from typing import Dict, Iterable
from hedra.versioning.flags.types.unstable.flag import unstable



class DistributionTypes(Enum):
    ALPHA='alpha'
    ANGLIT='anglit'
    ARCSINE='arcsine'
    ARGUS='argus'
    BETA_PRIME='beta-prime'
    BETA='beta'
    BRADFORD='bradford'
    BURR12='burr12'
    BURR='burr'
    CAUCHY='cauchy'
    CHI_SQUARED='chi-squared'
    CHI='chi'
    COSINE='cosine'
    CRYSTAL_BALL='crystal-ball'
    DGAMMA='dgamma'
    DWEIBULL='dweibull'
    ERLANG='erlang'
    EXPONENTIAL_NORMAL='exponential-normal'
    EXPONENTIAL_POWER='exponential-power'
    EXPONENTIAL='exponential'
    F_DISTRIBUTION='f-distribution'
    FATIGUE_LIFE='fatigue-life'
    FISK='fisk'
    FOLDED_CAUCHY='folded-cauchy'
    FOLDED_NORMAL='folded-normal'
    GAMMA='gamma'
    GAUSS_HYPERGEOMETRIC='gauss-hypergeometric'
    GENERALIZED_EXPONENTIAL='generalized-exponential'
    GENERALIZED_EXTREME='generalized-extreme'
    GENERALIZED_GAMMA='generalized-gamma'
    GENERALIZED_HALF_LOGISTIC='generalized-half-logistic'
    GENERALIZED_HYPERBOLIC='generalized-hyperbolic'
    GENERALIZED_LOGISTIC='generalized-logistic'
    GENERALIZED_NORMAL='generalized-normal'
    GENERALIZED_PARETO='generalized-pareto'
    GENERALIZED_INVERSE_GAUSS='generalized-inverse-gauss'
    GIBRAT='gibrat'
    GOMPERTZ='gompertz'
    GUMBEL_L='gumbel-l'
    GUMBEL_R='gumbel-r'
    HALF_CAUCHY='half-cauchy'
    HALF_GENERALIZED_NORMAL='half-generalized-normal'
    HALF_LOGISTIC='half-logistic'
    HALF_NORMAL='half-normal'
    HYPERBOLIC_SECANT='hyperbolic-secant'
    INVERSE_GAMMA='inverse-gamma'
    INVERSE_WEIBULL='inverse-weibull'
    JOHNSON_SB='johnson-db'
    JOHNSON_SU='johnson-su'
    KAPPA_3='kappa-3'
    KAPPA_4='kappa-4'
    KS_ONE='ks-one'
    KS_TWO_BIGENERALIZED='ks-two-bigeneralized'
    KS_TWO='ks-two'
    LAPLACE_ASYMMETRIC='laplace-asymmetric'
    LAPLACE='laplace'
    LEVY_L='levy-l'
    LEVY_STABLE='levy-stable'
    LEVY='levy'
    LOG_GAMMA='log-gamma'
    LOG_LAPLAPCE='log-laplace'
    LOG_UNIFORM='log-uniform'
    LOGISTIC='logistic'
    LOMAX='lomax'
    MAXWELL='maxwell'
    MIELKE='mielke'
    MOYAL='moyal'
    NAKAGAMI='nakagami'
    NON_CENTRAL_F_DISTRIBUTION='non-centeral-f'
    NON_CENTRAL_CHI_SQUARED='non-central-chisqr'
    NON_CENTRAL_T_DISTRIBUTION='non-central-t'
    NORMAL_INVERSE_GAUSS='normal=inverse-gauss'
    NORMAL='normal'
    PARETO='pareto'
    PEARSON_3='pearson-3'
    POWER_LOG_NORMAL='power-log-normal'
    POWER_NORMAL='power-normal'
    POWERLAW='powerlaw'
    R_DISTRIBUTION='r-distribution'
    RAYLEIGH='rayleigh'
    RECIPROCAL_INVERSE_GAUSS='reciprocal-inverse-gauss'
    RICE='rice'
    SEMI_CIRCULAR='semi-circular'
    SKEWED_CAUCHY='skewed-cauchy'
    SKEWED_NORMAL='skewed-normal'
    STUDENT_RANGE='student-range'
    T_DISTRIBUTION='t-distribution'
    TRAPEZOID='trapezoid'
    TRIANGULAR='triangular'
    TRUNCATED_EXPONENTIAL='truncated-exponential'
    TRUNCATED_NORMAL='truncated-normal'
    TRUNCATED_PARETO='truncated-pareto'
    TRUNCATED_WEIBULL_MINIMUM='truncated-weibull-minimum'
    TUKEY_LAMBDA='tuke-lambda'
    UNIFORM='uniform'
    VONMISES_LINE='vonmises-line'
    VONMISES='vonmises'
    WALD='wald'
    WEIBULL_MAXIMUM='weibull-maximum'
    WEIBULL_MINIMUM='weibull-minimum'
    WRAPPED_CAUCHY='wrapped-cauchy'


@unstable
class DistributionMap:

    def __init__(self) -> None:
        self._types: Dict[str, DistributionTypes] = {
            distribution_type.value: distribution_type for distribution_type in DistributionTypes
        }

    def __iter__(self) -> Iterable[DistributionTypes]:
        for distribution_type in self._types.values():
            yield distribution_type

    def __getitem__(self, distribution_type: str) -> DistributionTypes:
        return self._types.get(
            distribution_type,
            DistributionTypes.NORMAL
        )

    def get(self, distribution_type: str) -> DistributionTypes:
        return self._types.get(
            distribution_type,
            DistributionTypes.NORMAL
        )


from typing import Dict, Union
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.action.event import ActionEvent
from hedra.core.hooks.types.channel.event import ChannelEvent
from hedra.core.hooks.types.check.event import CheckEvent
from hedra.core.hooks.types.condition.event import ConditionEvent
from hedra.core.hooks.types.context.event import ContextEvent
from hedra.core.hooks.types.event.event import Event
from hedra.core.hooks.types.load.event import LoadEvent
from hedra.core.hooks.types.metric.event import MetricEvent
from hedra.core.hooks.types.save.event import SaveEvent
from hedra.core.hooks.types.task.event import TaskEvent
from hedra.core.hooks.types.transform.event import TransformEvent


HedraEvent = Union[
    Event, 
    TransformEvent, 
    ConditionEvent, 
    ContextEvent, 
    SaveEvent, 
    LoadEvent, 
    CheckEvent
]


def get_event(target: Hook, source: Hook) -> HedraEvent:
    event_types: Dict[HookType, HedraEvent] = {
        HookType.ACTION: ActionEvent,
        HookType.CHANNEL: ChannelEvent,
        HookType.CHECK: CheckEvent,
        HookType.CONDITION: ConditionEvent,
        HookType.CONTEXT: ContextEvent,
        HookType.EVENT: Event,
        HookType.LOAD: LoadEvent,
        HookType.METRIC: MetricEvent,
        HookType.SAVE: SaveEvent,
        HookType.TASK: TaskEvent,
        HookType.TRANSFORM: TransformEvent
    }

    return event_types.get(
        source.hook_type, 
        Event
    )(target, source)



import uuid
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)

try:
    from hedra.reporting.types.statsd import StatsD
    from aio_statsd import TelegrafClient
    from .telegraf_config import TelegrafConfig
    has_connector = True

except Exception:
    from hedra.reporting.types.empty import Empty as StatsD
    TelegrafConfig=None
    has_connector = False


class Telegraf(StatsD):

    def __init__(self, config: TelegrafConfig) -> None:
        self.host = config.host
        self.port = config.port

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()


        self.connection = TelegrafClient(
            host=self.host,
            port=self.port
        )

        self.statsd_type = 'Telegraf'
        
    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to {self.statsd_type}')

        metrics_sets: List[SessionMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Preparing Session System Metrics Set - {metrics_set.system_metrics_set_id}')
  
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics.record)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics.record)

        for metrics_set in metrics_sets:

            for metric_field, metric_value in metrics_set.record.items():
                
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metric Set - {metrics_set.name}:{metrics_set.group}:{metric_field}')
                
                self.connection.send_telegraf(
                    f'{metrics_set.group}_{metrics_set.name}_{metric_field}',
                    metric_value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to {self.statsd_type}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to {self.statsd_type}')

        metrics_sets: List[SystemMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Preparing Stage System Metrics Set - {metrics_set.system_metrics_set_id}')
  
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics.record)

        for metrics_set in metrics_sets:

            for metric_field, metric_value in metrics_set.record.items():
                
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metric Set - {metrics_set.name}:{metrics_set.group}:{metric_field}')
                
                self.connection.send_telegraf(
                    f'{metrics_set.group}_{metrics_set.group}_{metrics_set.name}_{metric_field}',
                    metric_value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to {self.statsd_type}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to {self.statsd_type}')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():

                for metric_field, metric_value in group.items():
                    
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream Metric - {stage_name}:{group_name}:{metric_field}')
                    
                    self.connection.send_telegraf(
                        f'{stage_name}_stream_{group_name}_{metric_field}',
                        metric_value
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to {self.statsd_type}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to {self.statsd_type}')

        for experiment in experiment_metrics.experiment_summaries:

            experiment_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment - {experiment.experiment_name}:{experiment_id}')
            
            for field, value in experiment.stats:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment field - {experiment.experiment_name}:{field}')

                self.connection.send_telegraf(
                    f'{experiment.experiment_name}_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to {self.statsd_type}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to {self.statsd_type}')

        for variant in experiment_metrics.variant_summaries:

            variant_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variant - {variant.variant_name}:{variant_id}')
            
            for field, value in variant.stats:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants field - {variant.variant_name}:{field}')

                self.connection.send_telegraf(
                    f'{variant.variant_name}_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to {self.statsd_type}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to {self.statsd_type}')

        for mutation in experiment_metrics.mutation_summaries:

            mutation_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutation - {mutation.mutation_name}:{mutation_id}')
            
            for field, value in mutation.stats:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutatio field - {mutation.mutation_name}:{field}')

                self.connection.send_telegraf(
                    f'{mutation.mutation_name}_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to {self.statsd_type}')


    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to {self.statsd_type}')

        for event in events:
            self.connection.send_telegraf(event.name, {'time': event.time})
            
            if event.success:
                self.connection.send_telegraf(event.name, {'success': 1})
            
            else:
                self.connection.send_telegraf(event.name, {'failed': 1})

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to {self.statsd_type}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            self.connection.send_telegraf(
                f'{metrics_set.name}_common',
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to {self.statsd_type}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to {self.statsd_type}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')


            for group_name, group in metrics_set.groups.items():
                self.connection.send_telegraf(
                    f'{metrics_set.name}_{group_name}', 
                    {
                        **group.record,
                        'group': group_name
                    }
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to {self.statsd_type}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            self.connection.send_telegraf(
                f'{metrics_set.name}_custom',
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'custom',
                    **{
                        custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                    }
                }
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to {self.statsd_type}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                self.connection.send_telegraf(
                    f'{metrics_set.name}_errors', {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to {self.statsd_type}')


import json
import string
import random
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from hedra.core.experiments.mutations.types.base.mutation_type import MutationType
from typing import Union, Optional, Tuple
from .validator import InjectJunkDataValidator


Request = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action
]


class InjectJunkData(Mutation):

    def __init__(
        self, 
        name: str, 
        chance: float,
        *targets: Tuple[str, ...],
        junk_size: Optional[int]=None

    ) -> None:
        super().__init__(
            name, 
            chance,
            MutationType.INJECT_JUNK_DATA,
            *targets
        )

        validated_mutation = InjectJunkDataValidator(
            junk_size=junk_size
        )

        self.junk_size = validated_mutation.junk_size

        self._character_pool = ''.join([
            string.ascii_letters,
            string.digits,
            string.hexdigits,
            string.octdigits,
            string.punctuation,
            string.whitespace
        ])

        self.original_data: bytes = None

        junk_data: str = ''.join(
            random.choices(
                self._character_pool, 
                k=self.junk_size
            )
        )

        self.junk_data = junk_data.encode()

    async def mutate(self, action: Request=None) -> Request:

        chance_roll = random.uniform(0, 1)
        if chance_roll <= self.chance:
            return action

        if action.method not in ['POST', 'PUT', 'PATCH']:
            return action

        if self.original_data is None:
            self.original_data = action.encoded_data

        action.encoded_data = self.original_data + self.junk_data
        action._setup_data()
        action._setup_headers()
        
        return action
    
    def copy(self):
        return InjectJunkData(
            self.name,
            self.chance,
            *list(self.targets),
            junk_size=self.junk_size
        )


import os
from typing import Dict, List, Any
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.graphs.stages import Checkpoint
from hedra.core.hooks.types import context
from hedra.core.hooks.types import save


class CheckpointStage(Checkpoint):

    @save(save_path=f'{os.getcwd()}/checkpoint.json')
    async def save_results(
        self, 
        results: List[BaseResult]=[]
    ) -> List[Dict[str, Any]]:
        return [
            {
                'action_id': data.action_id,
                'elapsed': data.complete - data.start
            } for data in results
        ]


import asyncio
import time
import signal
from collections import defaultdict
from hedra.core.experiments.variant import Variant
from hedra.core.engines.client.config import Config
from hedra.core.personas.batching.param_type import ParamType
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.versioning.flags.types.unstable.flag import unstable
from statistics import mean
from typing import (
    Dict, 
    Any, 
    List, 
    Union, 
    Tuple
)
from .algorithms.types.point_optimizer import PointOptimizer
from .optimizer import Optimizer


@unstable
class DistributionFitOptimizer(Optimizer):
    
    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config)

        self.optimization_start = 0
        self.algorithm: PointOptimizer = None
        
        self.variant_weight = self.stage_config.experiment.get('weight')
        self.distribution_intervals = self.stage_config.experiment.get('intervals')
        self.distribution_type = self.stage_config.experiment.get('distribution_type')
        self.distribution = self.stage_config.experiment.get('distribution')

        self.variant = Variant(
            self.stage_name,
            weight=self.variant_weight,
            distribution=self.distribution_type,
        )

        self.algorithms: List[PointOptimizer] = []
        self.target_interval_completions: int = 0
        self.completion_rates = defaultdict(list)

        for distribution_idx in range(len(self.distribution)):
            self.algorithms.append(PointOptimizer(
                {
                    **config,
                    'stage_config': self.stage_config
                },
                distribution_idx=distribution_idx
            ))
            

    def _setup_persona(self, stage_config: Config) -> DefaultPersona:
        persona = DefaultPersona(stage_config)
        persona.setup(self.stage_hooks, self.metadata_string)

        return persona
    
    def optimize(self) -> Dict[str, Union[int, float]]:

        self._event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._event_loop)

        def handle_loop_stop(signame):
            try:
                self._event_loop.close()

            except BrokenPipeError:
                pass
                
            except RuntimeError:
                pass

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):

            self._event_loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(signame)
            )

        self._event_loop.set_exception_handler(self._handle_async_exception)

        optimization_results = self._event_loop.run_until_complete(
            self._optimize_async()
        )

        self._event_loop.close()

        return optimization_results

    
    async def _optimize_async(self) -> Dict[str, Union[int, float]]:

        results = None

        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Starting optimization')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Time Limit - {self._optimization_time_limit}')
        self.start = 0

        distribution_optimized_params = []
        distribution_mean_errors = []

        self.optimization_start = time.time()

        for distribution_value, algorithm in zip(self.distribution, self.algorithms):
            self.completion_rates = defaultdict(list)
            self.algorithm: PointOptimizer = algorithm
            self.algorithm.batch_time = self.stage_config.total_time/self.distribution_intervals

            self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Algorithm - {self.algorithm_type}')
            self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Batch Time - {self.algorithm.batch_time}')
            self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Max Iter - {self.algorithm.max_iter}')


            self.target_interval_completions = distribution_value

            self.elapsed = 0
            self._current_iter = 0
            self.start = time.time()

            results: Dict[str, Dict[str, Union[int, float]]] = await self.algorithm.optimize(self._optimize_iteration)

            optimization_results = results.get('batch_size')
            optimized_distribution_value = optimization_results.get('minimized_distribution_value')
            optimized_error_mean = optimization_results.get('minimized_error_mean')


            distribution_optimized_params.append(optimized_distribution_value)   
            distribution_mean_errors.append(optimized_error_mean)     
        
        self.total_optimization_time = time.time() - self.optimization_start

        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization took - {round(self.total_optimization_time, 2)} - seconds')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization - max actions per second - {self._max_aps}')

        self.optimized_results = {
            'optimize_target_stage': self.target_stage_name,
            'optimized_distribution': distribution_optimized_params,
            'optimization_mean_error': mean(distribution_mean_errors),
            'optimization_iters': self.algorithm.max_iter,
            'optimization_iter_duation': self.algorithm.batch_time,
            'optimization_total_time': self.total_optimization_time,
            'optimization_max_aps': self._max_aps
        }

        self.total_optimization_time = time.time() - self.start

        return self.optimized_results

    async def _optimize_iteration(self, xargs: List[Union[int, float]]) -> Tuple[Union[None, float], int]:

        if self.elapsed > self.algorithm.time_limit:
            return None, self.target_interval_completions

        persona = self._setup_persona(self.stage_config)

        for idx, param in enumerate(xargs):
            param_name = self.algorithm.param_names[idx]
            param = self.algorithm.param_values.get(param_name, {})
            param_type = param.get('type')

            if param_type == ParamType.INTEGER:
                xargs[idx] = int(xargs[idx])

            else:
                xargs[idx] = float(xargs[idx])

            param['value'] = xargs[idx]

            self.current_params[param_name] = xargs[idx]
            self.algorithm.current_params[param_name] = xargs[idx]

        persona.total_time = self.algorithm.batch_time
        persona = self.algorithm.update_params(persona)
        await persona.set_concurrency(persona.batch.size)

        await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter}')

        await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Size - {persona.batch.size}')
        await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Interval - {persona.batch.interval}')
        await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Gradient - {persona.batch.gradient}')

        completed_count = 0

        try:
            results = await persona.execute()
            completed_count = len([result for result in results if result.error is None])

        except Exception:
            pass

        elapsed = persona.end - persona.start

        await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - took - {round(elapsed, 2)} - seconds')
    
        self.completion_rates[persona.batch.size].append(completed_count)
        error = completed_count - self.target_interval_completions

        await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Target error- {error}')

        self._current_iter += 1
        self.elapsed = time.time() - self.start

        if completed_count < 1:
            return -(self.base_batch_size**2),  self.target_interval_completions
        
        return error, self.target_interval_completions



from typing import Dict, Any


class ProtobufRegistry:

    def __init__(self) -> None:
        self._protobufs: Dict[str, Any]

    def __getitem__(self, action_name : str) -> Any:
        return self._protobufs.get(action_name)

    def __setitem__(self, action_name: str, protobuf: Any):
        self._protobufs[action_name] = protobuf

    def get(self, action_name) -> Any:
        return self._protobufs.get(Any)


def make_protobuf_registry():
    return ProtobufRegistry()


protobuf_registry = make_protobuf_registry()


from .graphql_processed_result import GraphQLProcessedResult
from .graphql_http2_processed_result import GraphQLHTTP2ProcessedResult
from .grpc_processed_result import GRPCProcessedResult
from .http_processed_result import HTTPProcessedResult
from .http2_processed_result import HTTP2ProcessedResult
from .http3_processed_result import HTTP3ProcessedResult
from .playwright_processed_result import PlaywrightProcessedResult
from .task_processed_result import TaskProcessedResult
from .udp_processed_result import UDPProcessedResult
from .websocket_processed_result import WebsocketProcessedResult


from typing import Union
from scipy.stats import burr
from .base import BaseDistribution


class BurrDistribution(BaseDistribution):

    def __init__(
        self, 
        size: int, 
        c_value=2,
        d_value=1,
        center: Union[int, float]=0.5,
        randomness: Union[int, float]=0.25
    ) -> None:
        super().__init__(
            size=size,
            center=center,
            randomness=randomness,
            frozen_distribution=burr(
                c_value,
                d_value,
                center=center,
                randomness=randomness,
            )
        )


from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.data.serializers.serializer_types.common.metadata_serializer import MetadataSerializer
from typing import (
    List, 
    Dict, 
    Union, 
    Any
)


class BaseSerializer:

    def __init__(self) -> None:
        self.metadata_serializer = MetadataSerializer()

    def action_to_serializable(
        self, 
        action: BaseAction
    ) -> Dict[str, Union[str, List[str]]]:
        return {
            'action_id': action.action_id,
            'name': action.name,
            'metadata': self.metadata_serializer.serialize_metadata(
                action.metadata
            ),
        }
    
    def result_to_serializable(
        self,
        result: BaseResult
    ) -> Dict[str, Any]:
        return {
            'name': result.name,
            'error': str(result.error),
            'source': result.source,
            'user': result.user,
            'tags': result.tags,
            'type': result.type,
            'wait_start': float(result.wait_start),
            'start': float(result.start),
            'connect_end': float(result.connect_end),
            'write_end': float(result.write_end),
            'complete': float(result.complete),
            'checks': result.checks
        }


import os
from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool
)
from hedra.data.connectors.common.connector_type import ConnectorType


class XMLConnectorConfig(BaseModel):
    filepath: StrictStr
    file_mode: StrictStr='r'
    reporter_type: ConnectorType=ConnectorType.XML



from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class BigTableConfig(BaseModel):
    service_account_json_path: str
    instance_id: str
    events_table: str = 'events'
    metrics_table: str = 'metrics'
    experiments_table: str= 'experiments'
    streams_table: str='streams'
    system_metrics_table: str='system_metrics'
    reporter_type: ReporterTypes=ReporterTypes.BigTable




from __future__ import annotations
from functools import reduce
from typing import List, Any, Union
from typing import Any, AsyncIterable
from hedra.tools.helpers import (
    awaitable,
    wrap
)


class AsyncList:

    def __init__(self, data=[]):
        self.data = data

    def __getitem__(self, start: int, stop: int=None) -> Union[AsyncList, Any]:
        if stop:
            subset_list = self.data[start:stop]
            return AsyncList(subset_list)

        return self.data[start]

    def __setitem__(self, index: int, value: Any) -> None:
        self.data[index] = value

    def __add__(self, value: Any) -> List[Any]:
        appended_list = [*self.data, value]
        return AsyncList(appended_list)

    def __mul__(self, value: int) -> AsyncList:
        multiplied_list = self.data * value
        return AsyncList(multiplied_list)

    def __rmul__(self, value: int) -> AsyncList:
        multiplied_list = self.data * value
        return AsyncList(multiplied_list)

    def __iter__(self) -> AsyncIterable:
        for item in self.data:
            yield item

    async def __aiter__(self) -> AsyncIterable:
        for item in self.data:
            yield item

    async def enum(self) -> AsyncIterable:
        for idx, item in enumerate(self.data):
            yield idx, item

    async def size(self) -> int:
        return await awaitable(len, self.data)

    async def is_empty(self) -> bool:
        list_size = await self.size()
        return list_size == 0

    async def append(self, value) -> None:
        self.data = [
            *self.data,
            value
        ]

    async def prepend(self, value) -> None:
        self.data = [
            value,
            self.data
        ]

    async def insert(self, idx, value) -> None:
        size = await self.size(self.data)
        if idx >= size:
            await self.append(value)

        elif idx < 0:
            await self.prepend(value)

        else:
            self.data = [
                self.data[:idx-1],
                value,
                self.data[idx:]
            ]

    async def at(self, idx) -> Any:
        size = await self.size(self.data)
        if idx >= size:
            return None
        
        return self.data[idx]

    async def find(self, value) -> Any:
        for item in self.data:
            if item == value:
                return value

        return None

    async def exists(self, value) -> bool:
        found = await self.find(value)
        return found is not None

    async def sort(self, **kwargs) -> AsyncList:
        sorted_list = await awaitable(sorted, self.data, **kwargs)
        return AsyncList(sorted_list)

    async def filter(self, condition) -> AsyncList:
        filtered = []
        for item in self.data:
            meets_condition = await awaitable(condition, item)
            if meets_condition:
                filtered = [
                    *filtered,
                    item
                ]

        return AsyncList(filtered)

    async def map(self, transform) -> AsyncList:
        mapped = []
        for item in self.data:
            mapped_value = await awaitable(transform, item)
            mapped = [*mapped, mapped_value]

        return AsyncList(mapped)

    async def subscribe(self, hook, *args, **kwargs) -> None:
        async_hook = await wrap(hook, args, kwargs)
        async for item in async_hook():
            self.data = [
                *self.data,
                item
            ]

    async def publish(self, hook) -> None:
        for item in self.data:
            await awaitable(hook, item)

    async def join(self, join_character) -> str:
        return await awaitable(self.data.join, join_character)

    async def replace(self, match_value: Any, replacement: Any, occurences: int=-1, in_place=True) -> Union[AsyncList, None]:
        if occurences == -1:
            occurences = await self.size()

        if in_place:
            await self._replace_in_place(match_value, replacement, occurences)
        else:
            return await self._replace(match_value, replacement, occurences)
            
    async def _replace_in_place(self, match_value: Any, replacement: Any, occurences: int):    
        found = 0
        for idx, item in enumerate(self.data):
                if item == match_value and found < occurences:
                    self.data[idx] = replacement
                    found += 1

    async def _replace(self, match_value: Any, replacement: Any, occurences: int):
        replaced_list = AsyncList(list(self.data))
            
        found = 0
        for idx, item in enumerate(self.data):
                if item == match_value and found < occurences:
                    replaced_list[idx] = replacement
                    found += 1

        return replaced_list

    async def index(self, match_item: Any, offset: int=0, end: int=None):
        try:
            return await awaitable(self.data.index, match_item, offset, end)
        except ValueError:
            return -1

    async def indexes(self, match_item: Any) -> List[int]:

        occurences = []
        for idx in range(self.data):
            if self.data[idx] == match_item:
                occurences = [
                    *occurences,
                    idx
                ]

        return occurences

    async def count(self, match_item: Any) -> int:
        return await awaitable(self.data.count, match_item)

    async def reverse(self) -> None:
        await awaitable(self.data.reverse)

    async def pop(self) -> Any:
        return await awaitable(self.data.pop)

    async def remove(self, value: int) -> None:
        await awaitable(self.data.remove, value)

    async def remove_index(self, index: int) -> AsyncList:
        removed = self.data[index:index+1]
        return AsyncList(removed)

    async def remove_range(self, start: int=0, stop: int=None) -> AsyncList:
        if stop is None:
            stop = await self.size()
            return AsyncList(self.data[start:])
        elif start == 0:
            return AsyncList(self.data[:stop])
        else:
            removed = self.data[:start] + self.data[stop:]
            return AsyncList(removed)

    async def copy(self) -> AsyncList:
        return AsyncList(list(self.data))

    async def extend(self, value: List[Any]) -> None:
        await awaitable(self.data.extend, value)

    async def splice(self, start: int=0, stop: int=None, step=1) -> AsyncList:
        spliced = []
        for idx in range(start, stop, step):
            spliced = [
                *spliced,
                self.data[idx]
            ]

        return AsyncList(spliced)

    async def subset(self, values: List[Any]) -> AsyncList:

        found = []
        for value in values:
            found_item = await self.find(value)
            found = [
                *found,
                found_item
            ]

        return AsyncList(found)

    async def reduce(self, reducer: function) -> AsyncList:
        reduced = await awaitable(reduce, reducer, self.data)
        return AsyncList(reduced)

    async def maximum(self, *args, **kwargs):
        return await awaitable(max, self.data, *args, **kwargs)

    async def minimum(self, *args, **kwargs):
        return await awaitable(min, self.data, *args, **kwargs)

    async def clear(self) -> None:
        await awaitable(self.data.clear)



from __future__ import annotations
from hedra.versioning.flags.types.unstable.flag import unstable
from typing import Optional, Dict, Union
from .tracing_types import (
    Request,
    Response,
    RequestHook,
    ResponseHook,
    TraceSignal,
    UrlFilter
)
from .url_filters import default_params_strip_filter


def skip_import(*args, **kwargs):
    pass

try:
    from opentelemetry import context as context_api
    from opentelemetry import trace
    from opentelemetry.instrumentation.utils import (
        _SUPPRESS_INSTRUMENTATION_KEY,
        http_status_to_status_code
    )
    from opentelemetry.propagate import inject
    from opentelemetry.semconv.trace import SpanAttributes
    from opentelemetry.trace import SpanKind,  get_tracer, Span
    from opentelemetry.trace.status import Status, StatusCode
    from opentelemetry.util.http import remove_url_credentials

except ImportError:
    context_api = object
    trace = object
    _SUPPRESS_INSTRUMENTATION_KEY=None
    http_status_to_status_code = skip_import
    inject = skip_import
    SpanAttributes = object
    SpanKind = object
    get_tracer = skip_import
    Span = object
    Status = object
    StatusCode = object
    remove_url_credentials = skip_import


__name__ = 'hedra'
__version__ = "0.7.12"

OpenTelemetryTracingConfig = Union[
    UrlFilter,
    RequestHook,
    ResponseHook,
    TraceSignal
]


@unstable
class Trace:
    """First-class used to trace requests launched via ClientSession objects."""

    __slots__ = (
        'tracer',
        'span',
        'token',
        'allowed_traces',
        'url_filter',
        'request_hook',
        'response_hook',
        'on_request_chunk_sent',
        'on_request_headers_sent',
        'on_response_chunk_received',
        'on_request_redirect',
        'on_request_data_sent',
        'on_response_data_received',
        'on_response_headers_received',
        'on_connection_queued_start',
        'on_connection_queued_end',
        'on_connection_create_start',
        'on_connection_create_end',
        'on_connection_reuse_connection',
        'on_dns_resolve_host_start',
        'on_dns_resolve_host_end',
        'on_dns_cache_hit',
        'on_dns_cache_miss',
        'on_task_start',
        'on_task_end',
        'on_task_error',
        '_context_active'
    )

    def __init__(
        self, 
        url_filter: Optional[UrlFilter]=None,
        request_hook: Optional[RequestHook]=None,
        response_hook: Optional[ResponseHook]=None,
        on_request_headers_sent: Optional[TraceSignal]=None,
        on_request_data_sent: Optional[TraceSignal]=None,
        on_request_chunk_sent: Optional[TraceSignal]=None,
        on_response_headers_received: Optional[TraceSignal]=None,
        on_response_data_received: Optional[TraceSignal]=None,
        on_response_chunk_received: Optional[TraceSignal]=None,
        on_request_redirect: Optional[TraceSignal]=None,
        on_connection_queued_start: Optional[TraceSignal]=None,
        on_connection_queued_end: Optional[TraceSignal]=None,
        on_connection_create_start: Optional[TraceSignal]=None,
        on_connection_create_end: Optional[TraceSignal]=None,
        on_connection_reuse_connection: Optional[TraceSignal]=None,
        on_dns_resolve_host_start: Optional[TraceSignal]=None,
        on_dns_resolve_host_end: Optional[TraceSignal]=None,
        on_dns_cache_hit: Optional[TraceSignal]=None,
        on_dns_cache_miss: Optional[TraceSignal]=None,
        on_task_start: Optional[TraceSignal]=None,
        on_task_end: Optional[TraceSignal]=None,
        on_task_error: Optional[TraceSignal]=None
    ) -> None:
    
    
        
        self.allowed_traces = [
            'on_request_chunk_sent',
            'on_request_headers_sent',
            'on_response_chunk_received',
            'on_request_redirect',
            'on_request_data_sent',
            'on_response_data_received',
            'on_response_headers_received',
            'on_connection_queued_start',
            'on_connection_queued_end',
            'on_connection_create_start',
            'on_connection_create_end',
            'on_connection_reuse_connection',
            'on_dns_resolve_host_start',
            'on_dns_resolve_host_end',
            'on_dns_cache_hit',
            'on_dns_cache_miss',
            'on_task_start',
            'on_task_end',
            'on_task_error'
        ]

        if url_filter is None:
            url_filter = default_params_strip_filter

        self.url_filter = url_filter

        self.request_hook: Union[RequestHook, None] = request_hook
        self.response_hook: Union[RequestHook, None] = response_hook

        self.on_request_headers_sent: TraceSignal = on_request_headers_sent
        self.on_request_data_sent: TraceSignal = on_request_data_sent
        self.on_request_chunk_sent: TraceSignal = on_request_chunk_sent
        self.on_request_redirect: TraceSignal = on_request_redirect
        self.on_response_headers_received = on_response_headers_received
        self.on_response_data_received = on_response_data_received
        self.on_response_chunk_received: TraceSignal = on_response_chunk_received

        self.on_connection_queued_start: TraceSignal = on_connection_queued_start
        self.on_connection_queued_end: TraceSignal = on_connection_queued_end
        self.on_connection_create_start: TraceSignal = on_connection_create_start
        self.on_connection_create_end: TraceSignal = on_connection_create_end
        self.on_connection_reuse_connection: TraceSignal = on_connection_reuse_connection

        self.on_dns_resolve_host_start: TraceSignal = on_dns_resolve_host_start
        self.on_dns_resolve_host_end: TraceSignal = on_dns_resolve_host_end
        self.on_dns_cache_hit: TraceSignal = on_dns_cache_hit
        self.on_dns_cache_miss: TraceSignal = on_dns_cache_miss

        self.on_task_start = on_task_start
        self.on_task_end = on_task_end
        self.on_task_error = on_task_error

        self.tracer = get_tracer(__name__, __version__, None)
        self.span: Span = None
        self.token: object = None
        self._context_active = False

    def trace_config_ctx(
        self,
        **kwargs: Dict[str, TraceSignal]
    ) -> Trace:
        return Trace(
            tracer=self.tracer,
            url_filter=self.url_filter,
            request_hook=self.request_hook,
            response_hook=self.response_hook,
            on_request_headers_sent=self.on_request_chunk_sent,
            on_request_data_sent=self.on_request_data_sent,
            on_request_chunk_sent=self.on_request_chunk_sent,
            on_request_redirect=self.on_request_redirect,
            on_response_headers_received=self.on_response_headers_received,
            on_response_data_received=self.on_response_data_received,
            on_response_chunk_received=self.on_response_chunk_received,
            on_connection_queued_start=self.on_connection_queued_start,
            on_connection_queued_end=self.on_connection_queued_end,
            on_connection_create_start=self.on_connection_create_start,
            on_connection_create_end=self.on_connection_create_end,
            on_connection_reuse_connection=self.on_connection_reuse_connection,
            on_dns_resolve_host_start=self.on_dns_resolve_host_start,
            on_dns_resolve_host_end=self.on_dns_resolve_host_end,
            on_dns_cache_hit=self.on_dns_cache_hit,
            on_dns_cache_miss=self.on_dns_cache_miss,
            on_task_start=self.on_task_start,
            on_task_end=self.on_task_end,
            on_task_error=self.on_task_error,
            **kwargs
        )
    
    def to_dict(self) -> Dict[str, OpenTelemetryTracingConfig]:
        return {
            'url_filter': self.url_filter,
            'request_hook': self.request_hook,
            'response_hook': self.response_hook,
            'on_request_headers_sent': self.on_request_chunk_sent,
            'on_request_data_sent': self.on_request_data_sent,
            'on_request_chunk_sent': self.on_request_chunk_sent,
            'on_request_redirect': self.on_request_redirect,
            'on_response_headers_received': self.on_response_headers_received,
            'on_response_data_received': self.on_response_data_received,
            'on_response_chunk_received': self.on_response_chunk_received,
            'on_connection_queued_start': self.on_connection_queued_start,
            'on_connection_queued_end': self.on_connection_queued_end,
            'on_connection_create_start': self.on_connection_create_start,
            'on_connection_create_end': self.on_connection_create_end,
            'on_connection_reuse_connection': self.on_connection_reuse_connection,
            'on_dns_resolve_host_start': self.on_dns_resolve_host_start,
            'on_dns_resolve_host_end': self.on_dns_resolve_host_end,
            'on_dns_cache_hit': self.on_dns_cache_hit,
            'on_dns_cache_miss': self.on_dns_cache_miss,
            'on_task_start': self.on_task_start,
            'on_task_end': self.on_task_end,
            'on_task_error': self.on_task_error,
        }
    
    def add_trace(self, trace_signal: TraceSignal) -> None:
        trace_name = trace_signal.__name__

        if trace_name in self.allowed_traces:
            object.__setattr__(
                self,
                trace_name,
                trace_signal
            )

    async def on_request_start(
        self,
        request: Request,
    ):
        if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
            self.span = None
            return

        http_method = request.method.upper()
        request_span_name = f"HTTP {http_method}"

        request_url = request.url.full

        if callable(self.url_filter):
            request_url = (
                remove_url_credentials(
                    self.url_filter(request.url)
                )
            )
        
        else:
            remove_url_credentials(request.url.full)

        span_attributes = {
            SpanAttributes.HTTP_METHOD: http_method,
            SpanAttributes.HTTP_URL: request_url,
        }

        self.span = self.tracer.start_span(
            request_span_name, kind=SpanKind.CLIENT, attributes=span_attributes
        )

        if callable(self.request_hook):
            self.request_hook(self.span, request)

        self.token = context_api.attach(
            trace.set_span_in_context(self.span)
        )

        self._context_active = True

        inject(request.headers)

    async def on_request_end(
        self,
        response: Response,
    ):
        if self.span is None:
            return

        if callable(self.response_hook):
            self.response_hook(self.span, response)

        if self.span.is_recording() and response.status:
            self.span.set_status(
                Status(
                    http_status_to_status_code(response.status)
                )
            )

            self.span.set_attribute(
                SpanAttributes.HTTP_STATUS_CODE, response.status
            )

        self._end_trace()

    async def on_request_exception(
        self,
        response: Response,
    ) -> None:
        if self.span is None:
            return

        if self.span.is_recording() and response.error:

            if response.status:
                self.span.set_status(
                    Status(StatusCode.ERROR)
                )

            self.span.record_exception(
                str(response.error)
            )

        if callable(self.response_hook):
            self.response_hook(self.span, response)

        self._end_trace()
    
    def _end_trace(self) -> None:
        # context = context_api.get_current()

        if self._context_active:
            context_api.detach(self.token)
            self._context_active = False

            self.span.end()



class TransitionMetadata:

    def __init__(
        self, 
        allow_multiple_edges: bool,
        is_valid: bool
    ) -> None:
        self.allow_multiple_edges = allow_multiple_edges
        self.is_valid = is_valid


import asyncio
import functools
from hedra.distributed.models.dns import DNSMessage
from hedra.distributed.discovery.dns.core.record import RecordType
from typing import Dict, Callable, Tuple, Optional




class Memoizer:
    def __init__(self):
        self.data: Dict[str, asyncio.Task] = {}

    def memoize_async(
        self, 
        key: Callable[
            [
                Tuple[
                    Optional[DNSMessage], 
                    str,
                    RecordType
                ]
            ],
            Tuple[
                str,
                RecordType
            ]

        ]=None
    ):
        data = self.data

        def wrapper(func):

            @functools.wraps(func)
            async def wrapped(*args, **kwargs):

                cache_key = ()
                if key:
                    cache_key = key


                task = data.get(cache_key)

                if task is None:

                    task = asyncio.create_task(
                        func(*args, **kwargs)
                    )

                    data[cache_key] = task

                    task.add_done_callback(
                        lambda _: self.clear(cache_key)
                    )

                return await task

            return wrapped

        return wrapper
    
    def clear(self, key: str):
        self.data.pop(key, None)



from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictBytes, 
    Json, 
    StrictInt,
    validator
)

from typing import Optional, Union


class SmuggleRequestValidator(BaseModel):
    request_size: Optional[StrictInt]
    smuggled_request: Optional[Union[StrictStr, StrictBytes, Json]]

    @validator('request_size')
    def validate_request_size(cls, val):

        if val is None:
            assert cls.smuggled_request is not None, "Field smuggled_request cannot be None if request_size is None"

        return val

    @validator('smuggled_request')
    def validate_smuggled_request(cls, val):

        if val is None:
            assert cls.request_size is not None, "Field request_size cannot be None if smuggled_request is None"

        return val


import asyncio
import functools
import signal
import psutil
import uuid
from typing import (
    List, 
    Dict, 
    Any,
    Coroutine
)
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .google_cloud_storage_connector_config import GoogleCloudStorageConnectorConfig


try:

    from google.cloud import storage
    has_connector = True

except Exception:
    storage = None
    has_connector = False



def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()

    except Exception:
        pass


class GoogleCloudStorageConnector:
    connector_type=ConnectorType.GCS

    def __init__(
        self, 
        config: GoogleCloudStorageConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        
        self.service_account_json_path = config.service_account_json_path
        self.stage = stage
        self.parser_config = parser_config

        self.bucket_namespace = config.bucket_namespace
        self.bucket_name = config.bucket_name
       

        self.credentials = None
        self.client = None

        self._bucket = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

        self.parser = Parser()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        self.client = storage.Client.from_service_account_json(self.service_account_json_path)

        self._bucket = await self._loop.run_in_executor(
            self._executor,
            self.client.create_bucket,
            self.bucket_name
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)
    
    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data()

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data()

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        blobs: List[storage.Blob] = await self._loop.run_in_executor(
            self._executor,        
            self._bucket.list_blobs
        )

        return await asyncio.gather(*[
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._bucket.get_blob,
                    blob.name
                )
            ) for blob in blobs
        ])
    
    async def close(self):

        await self._loop.run_in_executor(
            self._executor,
            self.client.close
        )

        self._executor.shutdown(wait=False, cancel_futures=True)



from __future__ import annotations
import struct
from typing import Dict, Iterable, Tuple
from .record_data import RecordData
from .record_types import RecordType
from .utils import (
    load_domain_name,
    pack_domain_name
)


class SRVRecordData(RecordData):
    '''A record'''

    def __init__(self, *args):
        super().__init__(
            RecordType.SRV,
            data=args
        )

        (
            priority,
            weight,
            port,
            hostname
        ) = args

        self.priority = priority
        self.weight = weight
        self.port = port
        self.hostname = hostname

    def __repr__(self):
        return '<%s-%s: %s:%s>' % (
            self.type_name, 
            self.priority,
            self.hostname, 
            self.port
        )


    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int
    ) -> Tuple[int, SRVRecordData]:
        
        priority, weight, port = struct.unpack(
            '!HHH', 
            data[cursor_position:cursor_position + 6]
        )

        cursor_position, hostname = load_domain_name(
            data, 
            cursor_position + 6
        )

        return cursor_position, SRVRecordData(
            priority, 
            weight, 
            port, 
            hostname
        )

    def dump(
        self, 
        names: Dict[str, int], 
        offset: int
    ) -> Iterable[bytes]:
        
        record_bytes = struct.pack(
            '!HHH', 
            self.priority, 
            self.weight, 
            self.port
        )

        domain_name = pack_domain_name(
            self.hostname, 
            names, 
            offset + 8
        )

        record_data = [
            record_bytes,
            domain_name
        ]

        for data in record_data:
            yield data




from __future__ import annotations
import asyncio
from ssl import SSLContext
from typing import Optional, Tuple
from hedra.core.engines.types.common.protocols import TCPConnection
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer
from hedra.core.engines.types.common.protocols.shared.constants import _DEFAULT_LIMIT


class HTTPConnection:

    __slots__ = (
        'dns_address',
        'port',
        'ssl',
        'ip_addr',
        'lock',
        'reader',
        'writer',
        'connected',
        'reset_connection',
        'pending',
        '_connection_factory'
    )

    def __init__(self, reset_connection: bool=False) -> None:
        self.dns_address: str = None
        self.port: int = None
        self.ssl: SSLContext = None
        self.ip_addr = None
        self.lock = asyncio.Lock()

        self.reader: Reader = None
        self.writer: Writer = None

        self.connected = False
        self.reset_connection = reset_connection
        self.pending = 0
        self._connection_factory = TCPConnection()

    async def make_connection(
        self, 
        hostname: str, 
        dns_address: str,
        port: int, 
        socket_config: Tuple[int, int, int, int, Tuple[int, int]],
        ssl: Optional[SSLContext]=None,
        timeout: Optional[float]=None
    ) -> None:
        if self.connected is False or self.dns_address != dns_address or self.reset_connection:
            try:
                reader, writer = await asyncio.wait_for(self._connection_factory.create(hostname, socket_config, ssl=ssl), timeout=timeout)
                self.connected = True

                self.reader = reader
                self.writer = writer

                self.dns_address = dns_address
                self.port = port
                self.ssl = ssl

            except asyncio.TimeoutError:
                raise Exception('Connection timed out.')

            except (ConnectionResetError, OSError,):
                raise Exception('Connection reset.')

            except Exception as e:
                raise e

    @property
    def empty(self):
        return not self.reader._buffer

    def read(self):
        return self.reader.read(n=_DEFAULT_LIMIT)

    def readexactly(self, n_bytes: int):
        return self.reader.read(n=n_bytes)

    def readuntil(self, sep=b'\n'):
        return self.reader.readuntil(separator=sep)

    def write(self, data):
        self.writer.write(data)

    def reset_buffer(self):
        self.reader._buffer = bytearray()

    def read_headers(self):
        return self.reader.read_headers()

    async def close(self):
        try:
            await self._connection_factory.close()
        except Exception:
            pass


import inspect
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.graphs.stages.base.exceptions.reserved_method_error import ReservedMethodError
from hedra.core.graphs.stages.base.stage import Stage


class Error(Stage):
    stage_type=StageTypes.ERROR

    def __init__(self) -> None:
        super().__init__()
        self.error = None
        self.retries: int = 0

        self.priority = None
        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        base_stage_name = self.__class__.__name__
        self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Checking internal Hooks for stage - {base_stage_name}')
        
        for reserved_hook_name in self.internal_hooks:
            try:

                hook = registrar.reserved[base_stage_name].get(reserved_hook_name)

                assert hasattr(self, reserved_hook_name) is True
                assert isinstance(hook, Hook) is True
                assert hook.hook_type == HookType.INTERNAL

                internal_hook = getattr(self, hook.shortname)
                assert inspect.getsource(internal_hook) == inspect.getsource(hook._call)

            except AssertionError:
                raise ReservedMethodError(self, reserved_hook_name)

            hook._call = hook._call.__get__(self, self.__class__)
            setattr(self, reserved_hook_name, hook._call)
            
            self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Loading internal Hook - {hook.name} - for stage - {base_stage_name}')

    @Internal()
    async def run(self):
        await self.logger.spinner.system.error(f'{self.metadata_string} - Encountered error - {self.error}')
        await self.logger.filesystem.aio['hedra.core'].error(f'{self.metadata_string} - Encountered error - {self.error}')
        





import asyncio
from contextlib import AbstractAsyncContextManager
from types import TracebackType
from typing import (
    Dict, 
    Optional, 
    Type
)



class BaseLimiter(AbstractAsyncContextManager):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop"
    )

    def __init__(
        self, 
        max_rate: float, 
        time_period: float = 60,
        reject_requests: bool=True
    ) -> None:
        self.max_rate = max_rate
        self.time_period = time_period
        self._rate_per_sec = max_rate / time_period
        self._level = 0.0

        self._waiters: Dict[asyncio.Task, asyncio.Future] = {}
        self._loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

        self._reject_requests = reject_requests

    def has_capacity(self, amount: float = 1) -> bool:
        raise NotImplementedError('Err. - has_capacity() is not implemented on BaseLimiter')

    async def acquire(
        self, 
        amount: float = 1,
    ):
        
        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")


        task = asyncio.current_task(
            loop=self._loop
        )

        assert task is not None

        rejected = False

        if not self.has_capacity(amount) and self._reject_requests:
            return True

        while not self.has_capacity(amount):

            fut = self._loop.create_future()
            try:

                self._waiters[task] = fut

                await asyncio.wait_for(
                    asyncio.shield(fut), 
                    timeout=(1 / self._rate_per_sec * amount)
                )

            except asyncio.TimeoutError:
                pass
                
            fut.cancel()

            if self._reject_requests:
                rejected = True

        self._waiters.pop(task, None)
        self._level += amount

        return rejected

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        return None


from .adaptive_limiter import AdaptiveRateLimiter
from .cpu_adaptive import CPUAdaptiveLimiter
from .leaky_bucket_limiter import LeakyBucketLimiter
from .resource_adaptive_limiter import ResourceAdaptiveLimiter
from .sliding_window_limiter import SlidingWindowLimiter
from .token_bucket_limiter import TokenBucketLimiter


from typing import Optional, Tuple
from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictInt,
    StrictBool
)


class ContextHookValidator(BaseModel):
    names: Optional[Tuple[StrictStr, ...]]
    order: StrictInt
    skip: StrictBool



import click
import os
from typing import (
    List, 
    Union, 
    Optional, 
    Callable, 
    Any
)
from importlib.metadata import version
from art import text2art
from hedra.logging import HedraLogger


class CLI(click.MultiCommand):

    command_files = {
        'ping': 'ping.py',
        'graph': 'graph.py',
        'project': 'project.py',
        'cloud': 'cloud.py',
        'plugin': 'plugin.py'
    }
    logger = HedraLogger()

    def __init__(
        self, 
        name: Optional[str] = None, 
        invoke_without_command: bool = False, 
        no_args_is_help: Optional[bool] = None, 
        subcommand_metavar: Optional[str] = None, 
        chain: bool = False, 
        result_callback: Optional[Callable[..., Any]] = None, 
        **attrs: Any
    ) -> None:
        super().__init__(
            name, 
            invoke_without_command, 
            no_args_is_help, 
            subcommand_metavar, 
            chain, 
            result_callback, 
            **attrs
        )

        self.logger.initialize()

        header_text = text2art(f'hedra', font='alligator').strip('\n')
        hedra_version = version('hedra')

        self.logger.console.sync.info(f'\n{header_text} {hedra_version}\n\n')

    def list_commands(self, ctx: click.Context) -> List[str]:
        rv = []
        for filename in self.command_files.values():
            rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx: click.Context, name: str) -> Union[click.Command, None]:

        ns = {}
        
        command_file = os.path.join(
            os.path.dirname(__file__),
            self.command_files.get(name, 'ping.py')
        )

        with open(command_file) as f:
            code = compile(f.read(), command_file, 'exec')
            eval(code, ns, ns)

        return ns.get(name)


from typing import Any
from hedra.versioning.flags.types.base.feature import Flag
from hedra.versioning.flags.types.base.flag_type import FlagTypes
from hedra.versioning.flags.exceptions.unsafe_not_enabled import UnsafeNotEnabledException


class UnsafeFeature(Flag):

    def __init__(
            self,
            feature_name: str,
            feature: Any
        ) -> None:
        super().__init__(
            feature_name,
            feature,
            FlagTypes.UNSAFE_FEATURE,
            UnsafeNotEnabledException
        )
        


from .client import MercuryPlaywrightClient
from .command import (
    PlaywrightCommand,
    Page,
    Input,
    URL,
    Options
)
from .result import PlaywrightResult

from .context_config import ContextConfig


import aiodns
import socket
from collections import defaultdict
from ipaddress import ip_address, IPv4Address
from urllib.parse import urlparse
from asyncio.events import get_event_loop
from typing import Dict, List
from .types import SocketProtocols, SocketTypes


class URL:

    __slots__ = (
        'resolver',
        'ip_addr',
        'parsed',
        'is_ssl',
        'port',
        'full',
        'has_ip_addr',
        'socket_config',
        'family',
        'protocol',
        'loop'
    )

    def __init__(self, url: str, port: int=80, family: SocketTypes=SocketTypes.DEFAULT, protocol: SocketProtocols=SocketProtocols.DEFAULT) -> None:
        self.resolver = aiodns.DNSResolver()
        self.ip_addr = None
        self.parsed = urlparse(url)
        self.is_ssl = 'https' in url or 'wss' in url   

        if self.is_ssl:
            port = 443

        self.port = self.parsed.port if self.parsed.port else port
        self.full = url
        self.has_ip_addr = False 
        self.socket_config = None
        self.family = family
        self.protocol = protocol
        self.loop = None

    async def replace(self, url: str):
        self.full = url
        self.params = urlparse(url)

    async def lookup(self):

        if self.loop is None:
            self.loop = get_event_loop()

        infos: Dict[str, List] = defaultdict(list)

        if self.parsed.hostname is None:
            
            try:

                address = self.full.split(':')
                assert len(address) == 2

                host, port = address
                self.port = int(port)

                if isinstance(ip_address(host), IPv4Address):
                    socket_type = socket.AF_INET

                else:
                    socket_type = socket.AF_INET6
                
                infos[self.full] = [(
                    socket_type,
                    self.protocol,
                    None,
                    None,
                    (
                        host,
                        self.port
                    )
                )]
            
            except Exception as parse_error:
                raise parse_error

        else:

            resolved = await self.resolver.gethostbyname(self.parsed.hostname, self.family)
    
            for address in resolved.addresses:

                if isinstance(ip_address(address), IPv4Address):
                    socket_type = socket.AF_INET

                else:
                    socket_type = socket.AF_INET6

                info = await self.loop.getaddrinfo(
                    address, 
                    self.port, 
                    family=self.family, 
                    proto=0, 
                    flags=0
                )

                if infos.get(address) is None:
                    infos[address] = [*info]

                else:
                    infos[address].extend(info)

        return infos

    @property
    def params(self):
        return self.parsed.params

    @params.setter
    def params(self, value: str):
        self.parsed = self.parsed._replace(params=value)

    @property
    def scheme(self):
        return self.parsed.scheme

    @scheme.setter
    def scheme(self, value):
        self.parsed = self.parsed._replace(scheme=value)

    @property
    def hostname(self):
        return self.parsed.hostname

    @hostname.setter
    def hostname(self, value):
        self.parsed = self.parsed._replace(hostname=value)

    @property
    def path(self):
        url_path = self.parsed.path
        url_query = self.parsed.query
        url_params = self.parsed.params

        if url_path and len(url_path) == 0:
            url_path = "/"

        if url_query and len(url_query) > 0:
            url_path += f'?{self.parsed.query}'

        elif url_params and len(url_params) > 0:
            url_path += self.parsed.params

        return url_path

    @path.setter
    def path(self, value):
        self.parsed = self.parsed._replace(path=value)
        
    @property
    def query(self):
        return self.parsed.query

    @query.setter
    def query(self, value):
        self.parsed = self.parsed._replace(query=value)

    @property
    def authority(self):
        return self.parsed.hostname

    @authority.setter
    def authority(self, value):
        self.parsed = self.parsed._replace(hostname=value)


import dill
import traceback
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql.result import GraphQLResult
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.graphql_http2.result import GraphQLHTTP2Result
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.grpc.result import GRPCResult
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http.result import HTTPResult
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http2.result import HTTP2Result
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.engines.types.http3.result import HTTP3Result
from hedra.core.engines.types.playwright.command import PlaywrightCommand
from hedra.core.engines.types.playwright.result import PlaywrightResult
from hedra.core.engines.types.task.task import Task
from hedra.core.engines.types.task.result import TaskResult
from hedra.core.engines.types.udp.action import UDPAction
from hedra.core.engines.types.udp.result import UDPResult
from hedra.core.engines.types.websocket.action import WebsocketAction
from hedra.core.engines.types.websocket.result import WebsocketResult
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.task.hook import TaskHook
from typing import Union, Callable, Dict, Any
from .serializer_types import (
    GraphQLSerializer,
    GraphQLHTTP2Serializer,
    GRPCSerializer,
    HTTPSerializer,
    HTTP2Serializer,
    HTTP3Serializer,
    PlaywrightSerializer,
    TaskSerializer,
    UDPSerializer,
    WebsocketSerializer
)


Action = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action,
    PlaywrightCommand,
    Task,
    UDPAction,
    WebsocketAction
]


Result = Union[
    GraphQLResult,
    GraphQLHTTP2Result,
    GRPCResult,
    HTTPResult,
    HTTP2Result,
    HTTP3Result,
    PlaywrightResult,
    TaskResult,
    UDPResult,
    WebsocketResult
]


class Serializer:

    def __init__(self) -> None:
        self._serializers: Dict[
            str,
            Callable[
                ...,
                Union[
                    GraphQLSerializer,
                    GraphQLHTTP2Serializer,
                    GRPCSerializer,
                    HTTPSerializer,
                    HTTP2Serializer,
                    HTTP3Serializer,
                    PlaywrightSerializer,
                    TaskSerializer,
                    UDPSerializer,
                    WebsocketSerializer
                ]
            ]
        ] = {
            RequestTypes.GRAPHQL: lambda: GraphQLSerializer(),
            RequestTypes.GRAPHQL_HTTP2: lambda: GraphQLHTTP2Serializer(),
            RequestTypes.GRPC: lambda: GRPCSerializer(),
            RequestTypes.HTTP: lambda: HTTPSerializer(),
            RequestTypes.HTTP2: lambda: HTTP2Serializer(),
            RequestTypes.HTTP3: lambda: HTTP3Serializer(),
            RequestTypes.PLAYWRIGHT: lambda: PlaywrightSerializer(),
            RequestTypes.TASK: lambda: TaskSerializer(),
            RequestTypes.UDP: lambda: UDPSerializer(),
            RequestTypes.WEBSOCKET: lambda: WebsocketSerializer()
        }

        self._active_serializers: Dict[
            str,
            Union[
                GraphQLSerializer,
                GraphQLHTTP2Serializer,
                GRPCSerializer,
                HTTPSerializer,
                HTTP2Serializer,
                HTTP3Serializer,
                PlaywrightSerializer,
                TaskSerializer,
                UDPSerializer,
                WebsocketSerializer
            ] 
        ] = {}

    def serialize_action(
        self,
        hook: Union[ActionHook, TaskHook]
    ):
        action: Action = hook.action
        serializer = self._active_serializers.get(action.type)

        if serializer is None and action.type in self._serializers:
            serializer = self._serializers.get(action.type)()
            self._active_serializers[action.type] = serializer

        serializable_action = serializer.action_to_serializable(action)
        serializable_hook = hook.to_dict()
        serializable_client_config = hook.session.config_to_dict()

        return dill.dumps({
            'hook': serializable_hook,
            'action': serializable_action,
            'client_config': serializable_client_config
        })
    
    def deserialize_action(
        self,
        serialized_hook: Union[str, bytes]
    ):
        deserialized_hook: Dict[str, Any] = dill.loads(serialized_hook)

        deserialized_hook_config = deserialized_hook.get('hook', {})
        deserialized_action: Dict[str, Any] = deserialized_hook.get('action', {})
        deserialized_client_config = deserialized_hook.get('client_config', {})

        action_type = deserialized_action.get('type', RequestTypes.HTTP)

        serializer = self._active_serializers.get(action_type)
    
        if serializer is None and action_type in self._serializers:
            serializer = self._serializers.get(action_type)()
            self._active_serializers[action_type] = serializer

        if action_type == RequestTypes.TASK:

            action_hook = TaskHook(
                deserialized_hook_config.get('name'),
                deserialized_hook_config.get('shortname'),
                None,
                *deserialized_hook_config.get('names', []),
                weight=deserialized_hook_config.get('weight'),
                order=deserialized_hook_config.get('order'),
                skip=deserialized_hook_config.get('skip'),
                metadata={
                    'user': deserialized_hook_config.get('user'),
                    'tags': deserialized_hook_config.get('tags')
                }
            )

            action = serializer.deserialize_task(deserialized_action)
        
        else:

            action_hook = ActionHook(
                deserialized_hook_config.get('name'),
                deserialized_hook_config.get('shortname'),
                None,
                *deserialized_hook_config.get('names', []),
                weight=deserialized_hook_config.get('weight'),
                order=deserialized_hook_config.get('order'),
                skip=deserialized_hook_config.get('skip'),
                metadata={
                    'user': deserialized_hook_config.get('user'),
                    'tags': deserialized_hook_config.get('tags')
                }
            )

            action = serializer.deserialize_action(deserialized_action)

        action_hook.action = action

        session = serializer.deserialize_client_config(deserialized_client_config)
        action_hook.session = session

        return action_hook
    
    def serialize_result(
        self,
        result: Result
    ):
        serializer = self._active_serializers.get(result.type)

        if serializer is None and result.type in self._serializers:
            serializer = self._serializers.get(result.type)()
            self._active_serializers[result.type] = serializer

        serializable = serializer.result_to_serializable(result)

        return dill.dumps(serializable)
    
    def deserialize_result(
        self,
        serialized_result: Union[str, bytes]
    ) -> Result:
        deserialized_result: Dict[str, Any] = dill.loads(serialized_result)

        result_type = deserialized_result.get('type', RequestTypes.HTTP)

        serializer = self._active_serializers.get(result_type)
    
        if serializer is None and result_type in self._serializers:
            serializer = self._serializers.get(result_type)()
            self._active_serializers[result_type] = serializer

        return serializer.deserialize_result(deserialized_result)


from typing import List
from .stages import (
    AnalyzeStage,
    OptimizeStage,
    SetupStage,
)
from .stages.execute import (
    ExecuteGraphQLStage,
    ExecuteGraphQLHttp2Stage,
    ExecuteHTTPStage,
    ExecuteHTTP2Stage,
    ExecuteHTTP3Stage,
    ExecutePlaywrightStage,
    ExecuteTaskStage,
    ExecuteUDPStage,
    ExecuteWebsocketStage
)

from .stages.submit import (
    SubmitAWSLambdaResultsStage,
    SubmitAWSTimestreamResultsStage,
    SubmitBigQueryResultsStage,
    SubmitBigTableResultsStage,
    SubmitCassandraResultsStage,
    SubmitCloudwatchResultsStage,
    SubmitCosmosDBResultsStage,
    SubmitCSVResultsStage,
    SubmitDatadogResultsStage,
    SubmitDogStatsDResultsStage,
    SubmitGoogleCloudStorageResultsStage,
    SubmitGraphiteResultsStage,
    SubmitHoneycombResultsStage,
    SubmitInfluxDBResultsStage,
    SubmitJSONResultsStage,
    SubmitKafkaResultsStage,
    SubmitMongoDBResultsStage,
    SubmitMySQLResultsStage,
    SubmitNetdataResultsStage,
    SubmitNewrelicResultsStage,
    SubmitPostgresResultsStage,
    SubmitPrometheusResultsStage,
    SubmitRedisResultsStage,
    SubmitS3ResultsStage,
    SubmitSnowflakeResultsStage,
    SubmitSQLiteResultsStage,
    SubmitStatsDResultsStage,
    SubmitTelegrafResultsStage,
    SubmitTelegrafStatsDResultsStage,
    SubmitTimescaleDBResultsStage
)

from hedra.core.hooks import depends
from hedra.core.hooks.types.base.registrar import registrar
from hedra.projects.generation.generator import Generator


class GraphGenerator(Generator):

    def __init__(self) -> None:
        super().__init__({
            'analyze': AnalyzeStage,
            'aws-lambda': SubmitAWSLambdaResultsStage,
            'aws-timestream': SubmitAWSTimestreamResultsStage,
            'big-query': SubmitBigQueryResultsStage,
            'big-table': SubmitBigTableResultsStage,
            'cassandra': SubmitCassandraResultsStage,
            'cloudwatch': SubmitCloudwatchResultsStage,
            'cosmosdb': SubmitCosmosDBResultsStage,
            'csv': SubmitCSVResultsStage,
            'datadog': SubmitDatadogResultsStage,
            'dogstatsd': SubmitDogStatsDResultsStage,
            'google-cloud-storage': SubmitGoogleCloudStorageResultsStage,
            'graphite': SubmitGraphiteResultsStage,
            'graphql': ExecuteGraphQLStage,
            'graphql-http2': ExecuteGraphQLHttp2Stage,
            'honeycomb': SubmitHoneycombResultsStage,
            'http':ExecuteHTTPStage,
            'http2': ExecuteHTTP2Stage,
            'http3': ExecuteHTTP3Stage,
            'influxdb': SubmitInfluxDBResultsStage,
            'json': SubmitJSONResultsStage,
            'kafka': SubmitKafkaResultsStage,
            'mongodb': SubmitMongoDBResultsStage,
            'mysql': SubmitMySQLResultsStage,
            'netdata': SubmitNetdataResultsStage,
            'newrelic': SubmitNewrelicResultsStage,
            'optimize': OptimizeStage,
            'playwright': ExecutePlaywrightStage,
            'postgres': SubmitPostgresResultsStage,
            'prometheus': SubmitPrometheusResultsStage,
            'redis': SubmitRedisResultsStage,
            's3': SubmitS3ResultsStage,
            'setup': SetupStage,
            'snowflake': SubmitSnowflakeResultsStage,
            'sqlite': SubmitSQLiteResultsStage,
            'statsd': SubmitStatsDResultsStage,
            'task': ExecuteTaskStage,
            'telegraf': SubmitTelegrafResultsStage,
            'telegraf-statsd': SubmitTelegrafStatsDResultsStage,
            'timescaledb': SubmitTimescaleDBResultsStage,
            'udp': ExecuteUDPStage,
            'websocket': ExecuteWebsocketStage,
            'depends': depends
        }, registrar.module_paths)

        self.valid_types = [
            'analyze',
            'checkpoint',
            'execute',
            'optimize',
            'setup',
            'submit',
            'teardown',
            'validate'
        ]

    def generate_graph(
        self, 
        stages: List[str],
        engine: str=None,
        reporter: str=None
    ):

        if engine not in self.generator_types:
            engine = 'http'

        if reporter not in self.generator_types:
            reporter = 'json'

        for stage in stages:

            generator_type = stage
            if stage == "execute":
                generator_type = engine

            elif stage == "submit":
                generator_type = reporter

            modules = self.gather_required_items(generator_type)

            self.collect_imports(generator_type, modules)

        self.collect_imports(
            None,
            {
                'depends': depends
            }
        )
        
        self.serialize_items() 

        serialized_imports = '\n'.join([
            *self.serialized_global_imports,
            *self.serialized_local_imports
        ])

        for idx, serialized_stage in enumerate(self.serialized_locals):

            if idx > 0:
                previous_stage_name = self.locals[idx-1].__name__
                self.serialized_locals[idx] = f'@depends({previous_stage_name})\n{serialized_stage}'

        return '\n\n'.join([
            serialized_imports,
            *self.serialized_locals
        ])



from .cors import Cors
from .crsf import CRSF

from .circuit_breaker import CircuitBreaker

from .compressor import (
    BidirectionalGZipCompressor,
    BidirectionalZStandardCompressor,
    GZipCompressor,
    ZStandardCompressor
)

from .decompressor import (
    BidirectionalGZipDecompressor,
    BidirectionalZStandardDecompressor,
    GZipDecompressor,
    ZStandardDecompressor
)


from typing import Optional
from hedra.reporting.types.common.types import ReporterTypes
from pydantic import BaseModel


class MongoDBConfig(BaseModel):
    host: str='localhost:27017'
    username: Optional[str]
    password: Optional[str]
    database: str='hedra'
    events_collection: str='events'
    metrics_collection: str='metrics'
    experiments_collection: str='experiment'
    streams_collection: str='streams'
    system_metrics_collection: str='system_metrics'
    reporter_type: ReporterTypes=ReporterTypes.MongoDB


class DNSError(Exception):
    errors = {
        1: 'Format error: bad request',
        2: 'Server failure: error occurred',
        3: 'Name error: not exist',
        4: 'Not implemented: query type not supported',
        5: 'Refused: policy reasons'
    }

    def __init__(self, code: int, message: str = None):
        message = self.errors.get(code,
                                  message) or 'Unknown reply code: %d' % code
        super().__init__(message)
        self.code = code



from typing import List
from pydantic import BaseModel, conlist
from hedra.reporting.types.common.types import ReporterTypes


class _CloudwatchTarget(BaseModel):
    arn: str
    id: str

class CloudwatchConfig(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str
    iam_role_arn: str
    schedule_rate: str=None
    events_rule: str='hedra-events'
    metrics_rule: str='hedra-metrics'
    experiments_rule: str='hedra-experiments'
    streams_rule: str='hedra-streams'
    system_metrics_rule: str='system_metrics'
    cloudwatch_targets: conlist(_CloudwatchTarget, min_length=1)
    aws_resource_arns: List[str]=[]
    cloudwatch_source: str='hedra'
    submit_timeout: int=60
    reporter_type: ReporterTypes=ReporterTypes.Cloudwatch


import functools
from typing import Optional
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from .validator import MetricHookValidator




@registrar(HookType.METRIC)
def metric(
    metric_type: str, 
    group: Optional[str]='user_metrics',
    order: int=1,
    skip: bool=False
):
    
    MetricHookValidator(
        metric_type=metric_type,
        group=group,
        order=order,
        skip=skip
    )
    
    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper


from typing import Any
from hedra.core.engines.types.graphql import GraphQLResult
from .http_processed_result import HTTPProcessedResult


class HTTP3ProcessedResult(HTTPProcessedResult):

    def __init__(
        self, 
        stage: str, 
        result: GraphQLResult
    ) -> None:
        super(HTTP3ProcessedResult, self).__init__(
            stage,
            result
        )


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    StrictBytes
)

from typing import List, Optional


class ResultValidator(BaseModel):
    error: Optional[StrictStr]
    body: Optional[StrictBytes]
    status: StrictInt
    reason: Optional[StrictStr]
    params: Optional[StrictStr]
    query: StrictStr
    wait_start: StrictFloat
    start: StrictFloat
    connect_end: StrictFloat
    write_end: StrictFloat
    complete: StrictFloat
    checks: Optional[List[StrictStr]]


import statistics
import numpy
from collections import defaultdict
from typing import Dict, Union
from .system_metrics_set_types import (
    CPUMonitorGroup,
    MemoryMonitorGroup,
    StageSystemMetricsGroup,
    SystemMetricsCollection,
    SystemMetricGroupType
)


class SystemMetricsGroup:

    def __init__(
            self, 
            metrics: Union[CPUMonitorGroup, MemoryMonitorGroup],
            metric_group: SystemMetricGroupType
        ) -> None:
        self.stage_metrics: StageSystemMetricsGroup = defaultdict(dict)

        self.raw_metrics: Union[CPUMonitorGroup, MemoryMonitorGroup] = metrics
        self.metrics_group = metric_group
        self.metrics: Dict[str, Dict[str, SystemMetricsCollection]] = defaultdict(dict)
        self._quantiles = [
            10,
            20,
            25,
            30,
            40,
            50,
            60,
            70,
            75,
            80,
            90,
            95,
            99
        ]

        self.visibility_filters: Dict[str, Dict[str, bool]] = defaultdict(dict)

    def __iter__(self):
        for stage_metrics in self.metrics.values():
            for monitor_metrics in stage_metrics.values():
                yield monitor_metrics

    def aggregate(self):

        for stage_name, metrics in self.raw_metrics.items():
             for monitor_name, monitor_metrics in metrics.collected.items():

                self.visibility_filters[stage_name][monitor_name] = metrics.visibility_filters[monitor_name]

                if self.metrics_group == SystemMetricGroupType.MEMORY:
                    metrics_data = [
                        round(
                            metric_value/(1024**3),
                            2
                        ) for metric_value in monitor_metrics
                    ]

                else:
                    metrics_data = monitor_metrics

                if len(metrics_data) > 0:

                    self.stage_metrics[stage_name][monitor_name] = metrics.stage_metrics[monitor_name]

                    self.metrics[stage_name][monitor_name] = SystemMetricsCollection(**{
                        'stage': stage_name,
                        'name': monitor_name,
                        'group': self.metrics_group.value,
                        'mean': statistics.mean(metrics_data),
                        'median': statistics.median(metrics_data),
                        'max': max(metrics_data),
                        'min': min(metrics_data),
                        'stdev': statistics.stdev(metrics_data),
                        'variance': statistics.variance(metrics_data),
                        **{
                            f'quantile_{quantile}th':  numpy.quantile(
                                metrics_data,
                                round(
                                    quantile/100,
                                    2
                                )
                            ) for quantile in self._quantiles
                        }
                    })



# -*- coding: utf-8 -*-
# flake8: noqa
from typing import Dict
from collections import deque
from functools import lru_cache
from .exceptions import InvalidTableIndex


@lru_cache(maxsize=4096)
def table_entry_size(name, value):
    """
    Calculates the size of a single entry

    This size is mostly irrelevant to us and defined
    specifically to accommodate memory management for
    lower level implementations. The 32 extra bytes are
    considered the "maximum" overhead that would be
    required to represent each entry in the table.

    See RFC7541 Section 4.1
    """
    return 32 + len(name) + len(value)


STATIC_TABLE = (
        (b':authority'                  , b''             ),  # noqa
        (b':method'                     , b'GET'          ),  # noqa
        (b':method'                     , b'POST'         ),  # noqa
        (b':path'                       , b'/'            ),  # noqa
        (b':path'                       , b'/index.html'  ),  # noqa
        (b':scheme'                     , b'http'         ),  # noqa
        (b':scheme'                     , b'https'        ),  # noqa
        (b':status'                     , b'200'          ),  # noqa
        (b':status'                     , b'204'          ),  # noqa
        (b':status'                     , b'206'          ),  # noqa
        (b':status'                     , b'304'          ),  # noqa
        (b':status'                     , b'400'          ),  # noqa
        (b':status'                     , b'404'          ),  # noqa
        (b':status'                     , b'500'          ),  # noqa
        (b'accept-charset'              , b''             ),  # noqa
        (b'accept-encoding'             , b'gzip, deflate'),  # noqa
        (b'accept-language'             , b''             ),  # noqa
        (b'accept-ranges'               , b''             ),  # noqa
        (b'accept'                      , b''             ),  # noqa
        (b'access-control-allow-origin' , b''             ),  # noqa
        (b'age'                         , b''             ),  # noqa
        (b'allow'                       , b''             ),  # noqa
        (b'authorization'               , b''             ),  # noqa
        (b'cache-control'               , b''             ),  # noqa
        (b'content-disposition'         , b''             ),  # noqa
        (b'content-encoding'            , b''             ),  # noqa
        (b'content-language'            , b''             ),  # noqa
        (b'content-length'              , b''             ),  # noqa
        (b'content-location'            , b''             ),  # noqa
        (b'content-range'               , b''             ),  # noqa
        (b'content-type'                , b''             ),  # noqa
        (b'cookie'                      , b''             ),  # noqa
        (b'date'                        , b''             ),  # noqa
        (b'etag'                        , b''             ),  # noqa
        (b'expect'                      , b''             ),  # noqa
        (b'expires'                     , b''             ),  # noqa
        (b'from'                        , b''             ),  # noqa
        (b'host'                        , b''             ),  # noqa
        (b'if-match'                    , b''             ),  # noqa
        (b'if-modified-since'           , b''             ),  # noqa
        (b'if-none-match'               , b''             ),  # noqa
        (b'if-range'                    , b''             ),  # noqa
        (b'if-unmodified-since'         , b''             ),  # noqa
        (b'last-modified'               , b''             ),  # noqa
        (b'link'                        , b''             ),  # noqa
        (b'location'                    , b''             ),  # noqa
        (b'max-forwards'                , b''             ),  # noqa
        (b'proxy-authenticate'          , b''             ),  # noqa
        (b'proxy-authorization'         , b''             ),  # noqa
        (b'range'                       , b''             ),  # noqa
        (b'referer'                     , b''             ),  # noqa
        (b'refresh'                     , b''             ),  # noqa
        (b'retry-after'                 , b''             ),  # noqa
        (b'server'                      , b''             ),  # noqa
        (b'set-cookie'                  , b''             ),  # noqa
        (b'strict-transport-security'   , b''             ),  # noqa
        (b'transfer-encoding'           , b''             ),  # noqa
        (b'user-agent'                  , b''             ),  # noqa
        (b'vary'                        , b''             ),  # noqa
        (b'via'                         , b''             ),  # noqa
        (b'www-authenticate'            , b''             ),  # noqa
    )  # noqa

def _build_static_table_mapping() -> Dict[bytes, bytes]:
    """
    Build static table mapping from header name to tuple with next structure:
    (<minimal index of header>, <mapping from header value to it index>).

    static_table_mapping used for hash searching.
    """
    static_table_mapping = {}
    for index, (name, value) in enumerate(STATIC_TABLE, 1):
        header_name_search_result = static_table_mapping.setdefault(name, (index, {}))
        header_name_search_result[1][value] = index
    return static_table_mapping


STATIC_TABLE_LENGTH = len(STATIC_TABLE)
STATIC_TABLE_MAPPING = _build_static_table_mapping()

class HeaderTable:
    """
    Implements the combined static and dynamic header table

    The name and value arguments for all the functions
    should ONLY be byte strings (b'') however this is not
    strictly enforced in the interface.

    See RFC7541 Section 2.3
    """
    #: Default maximum size of the dynamic table. See
    #:  RFC7540 Section 6.5.2.
    DEFAULT_SIZE = 4096

    #: Constant list of static headers. See RFC7541 Section
    #:  2.3.1 and Appendix A
    


    def __init__(self):
        self._maxsize = HeaderTable.DEFAULT_SIZE
        self._current_size = 0
        self.resized = False
        self.dynamic_entries = deque()

    @lru_cache(maxsize=4096)
    def get_by_index(self, index):
        """
        Returns the entry specified by index

        Note that the table is 1-based ie an index of 0 is
        invalid.  This is due to the fact that a zero value
        index signals that a completely unindexed header
        follows.

        The entry will either be from the static table or
        the dynamic table depending on the value of index.
        """
        index -= 1
        if index < STATIC_TABLE_LENGTH:
            return STATIC_TABLE[index]

        index -= STATIC_TABLE_LENGTH
        if index < len(self.dynamic_entries):
            return self.dynamic_entries[index]

    def add(self, name, value):
        """
        Adds a new entry to the table

        We reduce the table size if the entry will make the
        table size greater than maxsize.
        """
        # We just clear the table if the entry is too big
        size = table_entry_size(name, value)
        if size > self._maxsize:
            self.dynamic_entries.clear()
            self._current_size = 0
        else:
            # Add new entry
            self.dynamic_entries.appendleft((name, value))
            self._current_size += size
            self._shrink()

    def search(self, name, value):
        """
        Searches the table for the entry specified by name
        and value

        Returns one of the following:
            - ``None``, no match at all
            - ``(index, name, None)`` for partial matches on name only.
            - ``(index, name, value)`` for perfect matches.
        """
        partial = None

        header_name_search_result = STATIC_TABLE_MAPPING.get(name)
        if header_name_search_result:
            index = header_name_search_result[1].get(value)
            if index is not None:
                return index, name, value
            else:
                partial = (header_name_search_result[0], name, None)

        offset = STATIC_TABLE_LENGTH + 1
        for (i, (n, v)) in enumerate(self.dynamic_entries):
            if n == name:
                if v == value:
                    return i + offset, n, v
                elif partial is None:
                    partial = (i + offset, n, None)
        return partial

    @property
    def maxsize(self):
        return self._maxsize

    @maxsize.setter
    def maxsize(self, newmax):
        newmax = int(newmax)
        oldmax = self._maxsize
        self._maxsize = newmax
        self.resized = (newmax != oldmax)
        if newmax <= 0:
            self.dynamic_entries.clear()
            self._current_size = 0
        elif oldmax > newmax:
            self._shrink()

    def _shrink(self):
        """
        Shrinks the dynamic table to be at or below maxsize
        """
        cursize = self._current_size
        while cursize > self._maxsize:
            name, value = self.dynamic_entries.pop()
            cursize -= table_entry_size(name, value)
        self._current_size = cursize




import time
from hedra.distributed.discovery.dns.core.record.record_data_types import (
    RecordData
)
from hedra.distributed.discovery.dns.core.record import (
    Record, 
    RecordType
)
from typing import Dict, Iterable, Tuple


class CacheValue:
    def __init__(self):
        self.data: Dict[RecordData, Dict[Tuple[int, RecordData], Record]] = {}

    def check_ttl(self, record: Record):
        return record.ttl < 0 or record.timestamp + record.ttl >= time.time()

    def get(
        self, 
        record_type: RecordType
    ) -> Iterable[Record]:

        if record_type == RecordType.ANY:
            for qt in self.data.keys():
                yield from self.get(qt)
        
        results = self.data.get(record_type)
        if results is not None:

            keys = list(results.keys())
            for key in keys:
                record = results[key]

                if self.check_ttl(record):
                    yield record
                    
                else:
                    results.pop(key, None)

    def add(self, record: Record):
        if self.check_ttl(record):
            results = self.data.setdefault(record.record_type, {})
            results[record.data] = record

